from abc import ABC, abstractmethod
from typing import Optional

from contaxy.operations import (
    AuthOperations,
    ExtensionOperations,
    FileOperations,
    JobOperations,
    JsonDocumentOperations,
    ProjectOperations,
    SeedOperations,
    ServiceOperations,
    SystemOperations,
)
from contaxy.schema.extension import CORE_EXTENSION_ID
from contaxy.utils.state_utils import GlobalState, RequestState


class ComponentOperations(ABC):
    @property
    @abstractmethod
    def global_state(self) -> GlobalState:
        pass

    @property
    @abstractmethod
    def request_state(self) -> RequestState:
        pass

    @abstractmethod
    def get_project_manager(self) -> ProjectOperations:
        """Returns a Project Manager instance."""
        pass

    @abstractmethod
    def get_auth_manager(self) -> AuthOperations:
        """Returns an Auth Manager instance."""
        pass

    @abstractmethod
    def get_system_manager(self) -> SystemOperations:
        """Returns a System Manager instance."""
        pass

    @abstractmethod
    def get_extension_manager(self) -> ExtensionOperations:
        """Returns an Extension Manager instance."""
        pass

    @abstractmethod
    def get_json_db_manager(self) -> JsonDocumentOperations:
        """Returns a JSON DB Manager instance."""
        pass

    @abstractmethod
    def get_file_manager(
        self, extension_id: Optional[str] = CORE_EXTENSION_ID
    ) -> FileOperations:
        """Returns a File Manager instance.

        Depending on the provided `extenion_id`, this is either the configured core component
        or an initialized extension client.

        Args:
            extension_id: ID of the requested extension. Use `core` for the platform components.
        """
        pass

    @abstractmethod
    def get_job_manager(
        self, extension_id: Optional[str] = CORE_EXTENSION_ID
    ) -> JobOperations:
        """Returns a Job Manager instance.

        Depending on the provided `extenion_id`, this is either the configured core component
        or an initialized extension client.

        Args:
            extension_id: ID of the requested extension. Use `core` for the platform components.
        """
        pass

    def get_service_manager(
        self, extension_id: Optional[str] = CORE_EXTENSION_ID
    ) -> ServiceOperations:
        """Returns a Service Manager instance.

        Depending on the provided `extenion_id`, this is either the configured core component
        or an initialized extension client.

        Args:
            extension_id: ID of the requested extension. Use `core` for the platform components.
        """
        pass

    @abstractmethod
    def get_seed_manager(self) -> SeedOperations:
        pass



from datetime import datetime
from typing import Any, List, Optional

from fastapi import APIRouter, Depends, Path, Query, Response, status

from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.managers.extension import parse_composite_id
from contaxy.schema import (
    ExtensibleOperations,
    Job,
    JobInput,
    ResourceAction,
    Service,
    ServiceInput,
)
from contaxy.schema.auth import AccessLevel, TokenPurpose, TokenType
from contaxy.schema.deployment import JOB_ID_PARAM, SERVICE_ID_PARAM, ServiceUpdate
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    CREATE_RESOURCE_RESPONSES,
    GET_RESOURCE_RESPONSES,
    UPDATE_RESOURCE_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
)
from contaxy.schema.extension import EXTENSION_ID_PARAM
from contaxy.schema.project import PROJECT_ID_PARAM
from contaxy.schema.shared import (
    OPEN_URL_REDIRECT,
    RESOURCE_ID_REGEX,
    CoreOperations,
    ResourceActionExecution,
)
from contaxy.utils import auth_utils

service_router = APIRouter(
    tags=["services"],
    responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE},
)
job_router = APIRouter(
    tags=["jobs"],
    responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE},
)


@service_router.get(
    "/projects/{project_id}/services",
    operation_id=ExtensibleOperations.LIST_SERVICES.value,
    response_model=List[Service],
    summary="List project services.",
    status_code=status.HTTP_200_OK,
)
def list_services(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all services associated with the given project."""
    component_manager.verify_access(
        token, f"projects/{project_id}/services", AccessLevel.READ
    )

    # TODO: List all services from all extensions
    return component_manager.get_service_manager(extension_id).list_services(project_id)


@service_router.get(
    "/projects/{project_id}/services:suggest-config",
    operation_id=ExtensibleOperations.SUGGEST_SERVICE_CONFIG.value,
    response_model=ServiceInput,
    summary="Suggest service configuration.",
    status_code=status.HTTP_200_OK,
)
def suggest_service_config(
    project_id: str = PROJECT_ID_PARAM,
    container_image: str = Query(
        ..., description="Container image to use for suggestion."
    ),
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Suggests an input configuration based on the provided `container_image`.

    The suggestion is based on metadata extracted from the container image (e.g. labels)
    as well as suggestions based on previous project deployments with the same image.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services:suggest-config", AccessLevel.WRITE
    )

    return component_manager.get_service_manager(extension_id).suggest_service_config(
        project_id, container_image
    )


@service_router.get(
    "/projects/{project_id}/services/{service_id}",
    operation_id=ExtensibleOperations.GET_SERVICE_METADATA.value,
    response_model=Service,
    summary="Get service metadata.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_service_metadata(
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the metadata of a single service.

    The returned metadata might be filtered based on the permission level of the authenticated user.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services/{service_id}", AccessLevel.READ
    )

    service_id, extension_id = parse_composite_id(service_id)
    return component_manager.get_service_manager(extension_id).get_service_metadata(
        project_id, service_id
    )


@service_router.post(
    "/projects/{project_id}/services:deploy-actions",
    operation_id=ExtensibleOperations.LIST_DEPLOY_SERVICE_ACTIONS.value,
    response_model=List[ResourceAction],
    summary="List deploy service actions.",
    status_code=status.HTTP_200_OK,
)
def list_deploy_service_actions(
    service: ServiceInput,
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all available service deployment options (actions).

    The returned action IDs should be used when calling the [deploy_service](#services/deploy_service) operation.

    The action mechanism allows extensions to provide additional deployment options for a service based on the provided configuration. It works the following way:

    1. The user requests all available deployment options via the [list_deploy_service_actions](#services/list_deploy_service_actions) operation.
    2. The operation will be forwarded to all installed extensions that have implemented the [list_deploy_service_actions](#services/list_deploy_service_actions) operation.
    3. Extensions can run arbitrary code based on the provided service configuration and return a list of actions with self-defined action IDs.
    4. The user selects one of those actions and triggers the [deploy_service](#services/deploy_service) operation by providing the selected action ID. The `action_id` from an extension contains the extension ID.
    5. The operation is forwarded to the selected extension, which can run arbitrary code to deploy the service based on the provided configuration.
    6. The return value of the operation should be a `Service` object.

    The same action mechanism is also used for other type of actions on resources.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services:deploy-actions", AccessLevel.WRITE
    )

    # TODO: List alldeploy  actions from all extensions
    return component_manager.get_service_manager(
        extension_id
    ).list_deploy_service_actions(project_id, service)


@service_router.post(
    "/projects/{project_id}/services",
    operation_id=ExtensibleOperations.DEPLOY_SERVICE.value,
    response_model=Service,
    summary="Deploy a service.",
    status_code=status.HTTP_200_OK,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def deploy_service(
    service: ServiceInput,
    project_id: str = PROJECT_ID_PARAM,
    action_id: Optional[str] = Query(
        None,
        description="The action ID from the service deploy options.",
        regex=RESOURCE_ID_REGEX,
    ),
    wait: bool = Query(
        False,
        description="If true, the server waits for the service to be ready before sending a response.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deploy a service for the specified project.

    If no `action_id` is provided, the system will automatically select the best deployment option.

    Available deployment options (actions) can be requested via the [list_deploy_service_actions](#services/list_deploy_service_actions) operation.
    If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

    The action mechanism is further explained in the description of the [list_deploy_service_actions](#services/list_deploy_service_actions).
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services", AccessLevel.WRITE
    )

    extension_id = None
    if action_id:
        action_id, extension_id = parse_composite_id(action_id)
    return component_manager.get_service_manager(extension_id).deploy_service(
        project_id, service, action_id, wait=wait
    )


@service_router.delete(
    "/projects/{project_id}/services/{service_id}",
    operation_id=ExtensibleOperations.DELETE_SERVICE.value,
    summary="Delete a service.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_service(
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    delete_volumes: Optional[bool] = Query(
        False, description="Delete all volumes associated with the deployment."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes a service.

    This will kill and remove the container and all associated deployment artifacts.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services/{service_id}", AccessLevel.WRITE
    )

    if not delete_volumes:
        delete_volumes = False

    service_id, extension_id = parse_composite_id(service_id)
    component_manager.get_service_manager(extension_id).delete_service(
        project_id, service_id, delete_volumes
    )
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@service_router.patch(
    "/projects/{project_id}/services/{service_id}",
    operation_id=ExtensibleOperations.UPDATE_SERVICE.value,
    summary="Update a service.",
    status_code=status.HTTP_200_OK,
    response_model=Service,
    responses={**UPDATE_RESOURCE_RESPONSES},
)
def update_service(
    service: ServiceUpdate,
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Updates a service.

    This might force a recreation of the underlying container.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services/{service_id}", AccessLevel.WRITE
    )
    service_id, extension_id = parse_composite_id(service_id)
    return component_manager.get_service_manager(extension_id).update_service(
        project_id, service_id, service
    )


@service_router.post(
    "/projects/{project_id}/services/{service_id}:update-service-access",
    operation_id=ExtensibleOperations.UPDATE_SERVICE_ACCESS.value,
    summary="Update a service.",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={**UPDATE_RESOURCE_RESPONSES},
)
def update_service_access(
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Update the last access information of the service to the current time and calling user."""
    component_manager.verify_access(
        token, f"projects/{project_id}/services/{service_id}", AccessLevel.WRITE
    )
    service_id, extension_id = parse_composite_id(service_id)
    component_manager.get_service_manager(extension_id).update_service_access(
        project_id, service_id
    )


@service_router.delete(
    "/projects/{project_id}/services",
    operation_id=ExtensibleOperations.DELETE_SERVICES.value,
    summary="Delete all services.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_services(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes all services associated with a project."""
    component_manager.verify_access(
        token, f"projects/{project_id}/services", AccessLevel.ADMIN
    )

    component_manager.get_service_manager(extension_id).delete_services(project_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@service_router.get(
    "/projects/{project_id}/services/{service_id}/logs",
    operation_id=ExtensibleOperations.GET_SERVICE_LOGS.value,
    response_model=str,
    summary="Get service logs.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_service_logs(
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    lines: Optional[int] = Query(None, description="Only show the last n lines."),
    since: Optional[datetime] = Query(
        None, description="Only show the logs generated after a given date."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the stdout/stderr logs of the service."""
    component_manager.verify_access(
        token, f"projects/{project_id}/services/{service_id}/logs", AccessLevel.WRITE
    )

    service_id, extension_id = parse_composite_id(service_id)
    return component_manager.get_service_manager(extension_id).get_service_logs(
        project_id, service_id, lines, since
    )


@service_router.get(
    "/projects/{project_id}/services/{service_id}/actions",
    operation_id=ExtensibleOperations.LIST_SERVICE_ACTIONS.value,
    response_model=List[ResourceAction],
    summary="List service actions.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def list_service_actions(
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all actions available for the specified service.

    The returned action IDs should be used when calling the [execute_service_action](#services/execute_service_action) operation.

    The action mechanism allows extensions to provide additional functionality on services. It works the following way:

    1. The user requests all available actions via the [list_service_actions](#services/list_service_actions) operation.
    2. The operation will be forwarded to all installed extensions that have implemented the [list_service_actions](#services/list_service_actions) operation.
    3. Extensions can run arbitrary code - e.g., request and check the service metadata for compatibility - and return a list of actions with self-defined action IDs.
    4. The user selects one of those actions and triggers the [execute_service_action](#services/execute_service_action) operation by providing the selected action ID.  The `action_id` from an extension contains the extension ID.
    5. The operation is forwarded to the selected extension, which can run arbitrary code to execute the selected action.
    6. The return value of the operation can be either a simple message (shown to the user) or a redirect to another URL (e.g., to show a web UI).

    The same action mechanism is also used for other resources (e.g., files, jobs).
    It can support a very broad set of use-cases, for example: Access to service endpoints, dashboards for monitoring, administration tools, and more...
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/services/{service_id}/actions", AccessLevel.WRITE
    )

    # TODO: List all actions from all extensions
    return component_manager.get_service_manager(extension_id).list_service_actions(
        project_id, service_id
    )


@service_router.post(
    "/projects/{project_id}/services/{service_id}/actions/{action_id}",
    operation_id=ExtensibleOperations.EXECUTE_SERVICE_ACTION.value,
    # TODO: what is the response model? add additional status codes?
    summary="Execute a service action.",
    status_code=status.HTTP_200_OK,
    responses={**OPEN_URL_REDIRECT, **GET_RESOURCE_RESPONSES},
)
def execute_service_action(
    action_execution: Optional[ResourceActionExecution] = None,
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    action_id: str = Path(
        ...,
        description="The action ID from the list_service_actions operation.",
        regex=RESOURCE_ID_REGEX,
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Executes the selected service action.

    The actions need to be first requested from the [list_service_actions](#services/list_service_actions) operation.
    If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

    The action mechanism is further explained in the description of the [list_service_actions](#services/list_service_actions).
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/services/{service_id}/actions/{action_id}",
        AccessLevel.WRITE,
    )

    action_id, extension_id = parse_composite_id(action_id)
    return component_manager.get_service_manager(extension_id).execute_service_action(
        project_id, service_id, action_id, action_execution or ResourceActionExecution()
    )


# Job Endpoints


@job_router.get(
    "/projects/{project_id}/jobs",
    operation_id=ExtensibleOperations.LIST_JOBS.value,
    response_model=List[Job],
    summary="List project jobs.",
    status_code=status.HTTP_200_OK,
)
def list_jobs(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all jobs associated with the given project."""
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs",
        AccessLevel.READ,
    )

    # TODO: List all jobs from all extensions
    return component_manager.get_job_manager(extension_id).list_jobs(project_id)


@job_router.delete(
    "/projects/{project_id}/jobs",
    operation_id=ExtensibleOperations.DELETE_JOBS.value,
    summary="Delete all jobs.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_jobs(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes all jobs associated with a project."""
    component_manager.verify_access(
        token, f"projects/{project_id}/jobs", AccessLevel.ADMIN
    )
    component_manager.get_job_manager(extension_id).delete_jobs(project_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@job_router.get(
    "/projects/{project_id}/jobs/{job_id}",
    operation_id=ExtensibleOperations.GET_JOB_METADATA.value,
    response_model=Job,
    summary="Get job metadata.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_job_metadata(
    project_id: str = PROJECT_ID_PARAM,
    job_id: str = JOB_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the metadata of a single job.

    The returned metadata might be filtered based on the permission level of the authenticated user.
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs/{job_id}",
        AccessLevel.READ,
    )

    job_id, extension_id = parse_composite_id(job_id)
    return component_manager.get_job_manager(extension_id).get_job_metadata(
        project_id, job_id
    )


@job_router.get(
    "/projects/{project_id}/jobs:suggest-config",
    operation_id=ExtensibleOperations.SUGGEST_JOB_CONFIG.value,
    response_model=JobInput,
    summary="Suggest job configuration.",
    status_code=status.HTTP_200_OK,
)
def suggest_job_config(
    project_id: str = PROJECT_ID_PARAM,
    container_image: str = Query(
        ..., description="Container image to use for suggestion."
    ),
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Suggests an input configuration based on the provided `container_image`.

    The suggestion is based on metadata extracted from the container image (e.g. labels)
    as well as suggestions based on previous project deployments with the same image.
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs:suggest-config",
        AccessLevel.WRITE,
    )

    return component_manager.get_job_manager(extension_id).suggest_job_config(
        project_id, container_image
    )


@job_router.post(
    "/projects/{project_id}/jobs:deploy-actions",
    operation_id=ExtensibleOperations.LIST_DEPLOY_JOB_ACTIONS.value,
    response_model=List[ResourceAction],
    summary="List deploy job actions.",
    status_code=status.HTTP_200_OK,
)
def list_deploy_job_actions(
    job: JobInput,
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all available job deployment options (actions).

    The returned action IDs should be used when calling the [deploy_job](#job/deploy_job) operation.

    The action mechanism allows extensions to provide additional deployment options for a job based on the provided configuration. It works the following way:

    1. The user requests all available deployment options via the [list_deploy_job_actions](#jobs/list_deploy_job_actions) operation.
    2. The operation will be forwarded to all installed extensions that have implemented the [list_deploy_job_actions](#jobs/list_deploy_job_actions) operation.
    3. Extensions can run arbitrary code based on the provided job configuration and return a list of actions with self-defined action IDs.
    4. The user selects one of those actions and triggers the [deploy_job](#jobs/deploy_job) operation by providing the selected action ID. The `action_id` from an extension contains the extension ID.
    5. The operation is forwarded to the selected extension, which can run arbitrary code to deploy the job based on the provided configuration.
    6. The return value of the operation should be a `Job` object.

    The same action mechanism is also used for other type of actions on resources.
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs:deploy-actions",
        AccessLevel.WRITE,
    )
    # TODO: List all deploy job actions from all extensions
    return component_manager.get_job_manager(extension_id).list_deploy_job_actions(
        project_id, job
    )


@job_router.post(
    "/projects/{project_id}/jobs",
    operation_id=ExtensibleOperations.DEPLOY_JOB.value,
    response_model=Job,
    summary="Deploy a job.",
    status_code=status.HTTP_200_OK,
    responses={**OPEN_URL_REDIRECT, **CREATE_RESOURCE_RESPONSES},
)
def deploy_job(
    job: JobInput,
    project_id: str = PROJECT_ID_PARAM,
    action_id: Optional[str] = Query(
        None,
        description="The action ID from the job deploy options.",
        regex=RESOURCE_ID_REGEX,
    ),
    wait: bool = Query(
        False,
        description="If true, the server waits for the job to be ready before sending a response.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deploy a job for the specified project.

    If no `action_id` is provided, the system will automatically select the best deployment option.

    Available actions can be requested via the [list_deploy_job_actions](#jobs/list_deploy_job_actions) operation.
    If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

    The action mechanism is further explained in the description of the [list_deploy_job_actions](#jobs/list_deploy_job_actions).
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs",
        AccessLevel.WRITE,
    )

    extension_id = None
    if action_id:
        action_id, extension_id = parse_composite_id(action_id)

    return component_manager.get_job_manager(extension_id).deploy_job(
        project_id, job, action_id, wait=wait
    )


@job_router.delete(
    "/projects/{project_id}/jobs/{job_id}",
    operation_id=ExtensibleOperations.DELETE_JOB.value,
    summary="Delete a job.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_job(
    project_id: str = PROJECT_ID_PARAM,
    job_id: str = JOB_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes a job.

    This will kill and remove the container and all associated deployment artifacts.
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs/{job_id}",
        AccessLevel.WRITE,
    )

    job_id, extension_id = parse_composite_id(job_id)
    component_manager.get_job_manager(extension_id).delete_job(project_id, job_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@job_router.get(
    "/projects/{project_id}/jobs/{job_id}/logs",
    operation_id=ExtensibleOperations.GET_JOB_LOGS.value,
    response_model=str,
    summary="Get job logs.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_job_logs(
    project_id: str = PROJECT_ID_PARAM,
    job_id: str = JOB_ID_PARAM,
    lines: Optional[int] = Query(None, description="Only show the last n lines."),
    since: Optional[datetime] = Query(
        None, description="Only show the logs generated after a given date."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the stdout/stderr logs of the job."""
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs/{job_id}/logs",
        AccessLevel.WRITE,
    )

    job_id, extension_id = parse_composite_id(job_id)
    return component_manager.get_job_manager(extension_id).get_job_logs(
        project_id, job_id, lines, since
    )


@job_router.get(
    "/projects/{project_id}/jobs/{job_id}/actions",
    operation_id=ExtensibleOperations.LIST_JOB_ACTIONS.value,
    response_model=List[ResourceAction],
    summary="List job actions.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def list_job_actions(
    project_id: str = PROJECT_ID_PARAM,
    job_id: str = JOB_ID_PARAM,
    extension_id: Optional[str] = EXTENSION_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all actions available for the specified job.

    The returned action IDs should be used when calling the [execute_job_action](#jobs/execute_job_action) operation.

    The action mechanism allows extensions to provide additional functionality on jobs. It works the following way:

    1. The user requests all available actions via the [list_job_actions](#jobs/list_job_actions) operation.
    2. The operation will be forwarded to all installed extensions that have implemented the [list_job_actions](#jobs/list_job_actions) operation.
    3. Extensions can run arbitrary code - e.g., request and check the job metadata for compatibility - and return a list of actions with self-defined action IDs.
    4. The user selects one of those actions and triggers the [execute_job_action](#jobs/execute_job_action) operation by providing the selected action ID. The `action_id` from an extension contains the extension ID.
    5. The operation is forwarded to the selected extension, which can run arbitrary code to execute the selected action.
    6. The return value of the operation can be either a simple message (shown to the user) or a redirect to another URL (e.g., to show a web UI).

    The same action mechanism is also used for other resources (e.g., files, services).
    It can support a very broad set of use-cases, for example: Access to dashboards for monitoring, adminsitration tools, and more...
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs/{job_id}/actions",
        AccessLevel.WRITE,
    )

    # TODO: List all job actions from all extensions

    return component_manager.get_job_manager(extension_id).list_job_actions(
        project_id, job_id
    )


@job_router.post(
    "/projects/{project_id}/jobs/{job_id}/actions/{action_id}",
    operation_id=ExtensibleOperations.EXECUTE_JOB_ACTION.value,
    # TODO: what is the response model? add additional status codes?
    summary="Execute a job action.",
    status_code=status.HTTP_200_OK,
    responses={**OPEN_URL_REDIRECT, **GET_RESOURCE_RESPONSES},
)
def execute_job_action(
    action_execution: Optional[ResourceActionExecution] = None,
    project_id: str = PROJECT_ID_PARAM,
    job_id: str = JOB_ID_PARAM,
    action_id: str = Path(
        ...,
        description="The action ID from the list_job_actions operation.",
        regex=RESOURCE_ID_REGEX,
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Executes the selected job action.

    The actions need to be first requested from the [list_job_actions](#jobs/list_job_actions) operation.
    If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

    The action mechanism is further explained in the description of the [list_job_actions](#jobs/list_job_actions).
    """
    component_manager.verify_access(
        token,
        f"projects/{project_id}/jobs/{job_id}/actions/{action_id}",
        AccessLevel.WRITE,
    )

    action_id, extension_id = parse_composite_id(action_id)
    return component_manager.get_job_manager(extension_id).execute_job_action(
        project_id, job_id, action_id, action_execution or ResourceActionExecution()
    )


@service_router.get(
    "/projects/{project_id}/services/{service_id}/token",
    operation_id=CoreOperations.GET_SERVICE_ACCESS_TOKEN.value,
    response_model=str,
    summary="Get service access token.",
    tags=["services"],
    status_code=status.HTTP_200_OK,
)
def get_service_access_token(
    project_id: str = PROJECT_ID_PARAM,
    service_id: str = SERVICE_ID_PARAM,
    endpoint: Optional[str] = Query(
        None, description="If specified, the token only allows access to this endpoint."
    ),
    description: Optional[str] = Query(
        None, description="Attach a short description to the generated token."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns an API token with permission to access the service endpoints.

    This token is read-only (permission level read) and does not allow any other permission such as deleting or updating the service.
    The token can be deleted (revoked) at any time.
    """
    authorized_access = component_manager.verify_access(
        token,
        f"projects/{project_id}/services/{service_id}/token",
        AccessLevel.WRITE,
    )

    if not endpoint:
        endpoint = ""

    access_service_scope = auth_utils.construct_permission(
        f"projects/{project_id}/services/{service_id}/access/{endpoint}",
        AccessLevel.READ,
    )

    component_manager.get_auth_manager().create_token(
        token_subject=authorized_access.authorized_subject,
        scopes=[access_service_scope],
        token_type=TokenType.API_TOKEN,
        description=description,
        token_purpose=TokenPurpose.SERVICE_ACCESS_TOKEN,
    )



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field


class JsonDocument(BaseModel):
    key: str = Field(
        ...,
        example="my-json-document",
        description="Unique key of the document.",
    )
    json_value: str = Field(
        ...,
        example="{'foo': 'bar'}",
        description="JSON value of the document.",
    )
    created_at: Optional[datetime] = Field(
        None,
        example="2021-04-23T10:20:30.400+02:30",
        description="Creation date of the document.",
    )
    created_by: Optional[str] = Field(
        None,
        example="16fd2706-8baf-433b-82eb-8c7fada847da",
        description="ID of the user that created this document.",
    )
    updated_at: Optional[datetime] = Field(
        None,
        example="2021-04-23T10:20:30.400+02:30",
        description="Last date at which the document was updated.",
    )
    updated_by: Optional[str] = Field(
        None,
        example="16fd2706-8baf-433b-82eb-8c7fada847da",
        description="ID of the user that has last updated this document.",
    )



from typing import Any, List, Optional

from fastapi import APIRouter, Depends, Query, Response, status
from fastapi.param_functions import Body

from contaxy import config
from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.schema import AccessLevel, CoreOperations, Project, ProjectInput, User
from contaxy.schema.auth import USER_ID_PARAM, TokenPurpose, TokenType
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    CREATE_RESOURCE_RESPONSES,
    GET_RESOURCE_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
    PermissionDeniedError,
)
from contaxy.schema.project import PROJECT_ID_PARAM, ProjectCreation
from contaxy.schema.shared import MAX_DISPLAY_NAME_LENGTH, MIN_DISPLAY_NAME_LENGTH
from contaxy.utils import auth_utils

router = APIRouter(
    tags=["projects"],
    responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE},
)


@router.post(
    "/projects",
    operation_id=CoreOperations.CREATE_PROJECT.value,
    response_model=Project,
    response_model_exclude_unset=True,
    summary="Create a new project.",
    status_code=status.HTTP_200_OK,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def create_project(
    project: ProjectCreation,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Creates a new project.

    We suggest to use the `suggest_project_id` endpoint to get a valid and available ID.
    The project ID might also be set manually, however, an error will be returned if it does not comply with the ID requirements or is already used.
    """
    # TODO: put method?

    # Check the token and get the authorized user
    authorized_access = component_manager.verify_access(token)
    # TODO: Check for permission in "project resource"
    # Check if the token has write access on the user
    component_manager.verify_access(
        token, authorized_access.authorized_subject, AccessLevel.WRITE
    )
    # Make sure the system project cannot be created via the API
    if project.id == config.SYSTEM_INTERNAL_PROJECT:
        raise PermissionDeniedError(f"The project id {project.id} is reserved.")
    return component_manager.get_project_manager().create_project(project)


@router.get(
    "/projects",
    operation_id=CoreOperations.LIST_PROJECTS.value,
    response_model=List[Project],
    response_model_exclude_unset=True,
    summary="List all projects.",
    status_code=status.HTTP_200_OK,
)
def list_projects(
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all projects visible to the authenticated user.

    A project is visible to a user, if the user has the atleast a `read` permission for the project.
    System adminstrators will also see technical projects, such as `system-internal` and `system-global`.
    """
    # Check the token and get the authorized user
    authorized_access = component_manager.verify_access(token)
    # TODO: Check for permission in "project resource"
    # Check if the token has read access on the user resource
    component_manager.verify_access(
        token, authorized_access.authorized_subject, AccessLevel.READ
    )
    return component_manager.get_project_manager().list_projects()


@router.patch(
    "/projects/{project_id}",
    operation_id=CoreOperations.UPDATE_PROJECT.value,
    response_model=Project,
    response_model_exclude_unset=True,
    summary="Update project metadata.",
    status_code=status.HTTP_200_OK,
)
def update_project(
    project_id: str = PROJECT_ID_PARAM,
    project: ProjectInput = Body(...),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Updates the metadata of the given project.

    This will update only the properties that are explicitly set in the patch request.
    The patching is based on the JSON Merge Patch Standard [RFC7396](https://tools.ietf.org/html/rfc7396).
    """
    component_manager.verify_access(token, f"projects/{project_id}", AccessLevel.WRITE)
    return component_manager.get_project_manager().update_project(project_id, project)


@router.get(
    "/projects/{project_id}",
    operation_id=CoreOperations.GET_PROJECT.value,
    response_model=Project,
    response_model_exclude_unset=True,
    summary="Get details for a project.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_project(
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the metadata of a single project."""
    component_manager.verify_access(token, f"projects/{project_id}", AccessLevel.READ)
    return component_manager.get_project_manager().get_project(project_id)


@router.get(
    "/projects:suggest-id",
    operation_id=CoreOperations.SUGGEST_PROJECT_ID.value,
    response_model=str,
    summary="Suggest project ID.",
    status_code=status.HTTP_200_OK,
)
def suggest_project_id(
    display_name: str = Query(
        ...,
        min_length=MIN_DISPLAY_NAME_LENGTH,
        max_length=MAX_DISPLAY_NAME_LENGTH,
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Suggests a valid and unique project ID for the given display name.

    The project ID will be human-readable and resemble the provided display name,
    but might be cut off or have an attached counter prefix.
    """
    # Only check the token for validity
    component_manager.verify_access(token)
    return component_manager.get_project_manager().suggest_project_id(display_name)


@router.delete(
    "/projects/{project_id}",
    operation_id=CoreOperations.DELETE_PROJECT.value,
    summary="Delete a project.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_project(
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes a project and all its associated resources including deployments and files.

    A project can only be delete by a user with `admin` permission on the project.
    """
    component_manager.verify_access(token, f"projects/{project_id}", AccessLevel.ADMIN)
    component_manager.get_project_manager().delete_project(project_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/projects/{project_id}/users",
    operation_id=CoreOperations.LIST_PROJECT_MEMBERS.value,
    summary="List project members.",
    response_model=List[User],
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
)
def list_project_members(
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all project members.

    This include all users that have atlease a `read` permission on the given project.
    """
    component_manager.verify_access(token, f"projects/{project_id}", AccessLevel.WRITE)
    return component_manager.get_project_manager().list_project_members(project_id)


@router.put(
    "/projects/{project_id}/users/{user_id}",
    operation_id=CoreOperations.ADD_PROJECT_MEMBER.value,
    summary="Add user to project.",
    response_model=List[User],
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
)
def add_project_member(
    project_id: str = PROJECT_ID_PARAM,
    user_id: str = USER_ID_PARAM,
    access_level: AccessLevel = Query(
        AccessLevel.WRITE,
        description="The permission level.",
        type="string",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Adds a user to the project.

    This will add the permission for this project to the user item.
    The `access_level` defines what the user can do:

    - The `read` permission level allows read-only access on all resources.
    - The `write` permission level allows to create and delete project resources.
    - The `admin` permission level allows to delete the project or add/remove other users.
    """
    component_manager.verify_access(token, f"projects/{project_id}", AccessLevel.ADMIN)
    return component_manager.get_project_manager().add_project_member(
        project_id, user_id, access_level
    )


@router.delete(
    "/projects/{project_id}/users/{user_id}",
    operation_id=CoreOperations.REMOVE_PROJECT_MEMBER.value,
    summary="Remove user from project.",
    response_model=List[User],
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
)
def remove_project_member(
    project_id: str = PROJECT_ID_PARAM,
    user_id: str = USER_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Removes a user from a project.

    This will remove the permission for this project from the user item.
    """
    component_manager.verify_access(token, f"projects/{project_id}", AccessLevel.ADMIN)
    return component_manager.get_project_manager().remove_project_member(
        project_id, user_id
    )


@router.get(
    "/projects/{project_id}/token",
    operation_id=CoreOperations.GET_PROJECT_TOKEN.value,
    response_model=str,
    summary="Get project token.",
    tags=["projects"],
    status_code=status.HTTP_200_OK,
)
def get_project_token(
    project_id: str = PROJECT_ID_PARAM,
    access_level: AccessLevel = Query(
        AccessLevel.WRITE,
        description="Access level of the token.",
        type="string",
    ),
    description: Optional[str] = Query(
        None, description="Attach a short description to the generated token."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns an API token with permission (`read`, `write`, or `admin`) to access all project resources.

    The `read` access level allows read-only access on all resources.
    The `write` access level allows to create and delete project resources.
    The `admin` access level allows to delete the project or add/remove other users.
    """
    access_level_to_check = access_level
    if access_level_to_check not in [AccessLevel.ADMIN, AccessLevel.WRITE]:
        # WRITE Access minimum should be the minimum to create tokens
        access_level_to_check = AccessLevel.WRITE

    authorized_access = component_manager.verify_access(
        token, f"projects/{project_id}", access_level_to_check
    )

    project_permission = auth_utils.construct_permission(
        f"projects/{project_id}", access_level
    )
    return component_manager.get_auth_manager().create_token(
        scopes=[project_permission],
        token_type=TokenType.API_TOKEN,
        token_subject=authorized_access.authorized_subject,
        token_purpose=TokenPurpose.PROJECT_API_TOKEN,
        description=description,
    )



from json.decoder import JSONDecodeError
from typing import Dict, List, Optional

import requests
from pydantic import parse_raw_as

from contaxy.clients.shared import handle_errors
from contaxy.operations import JsonDocumentOperations
from contaxy.schema import JsonDocument
from contaxy.schema.exceptions import ClientValueError


class JsonDocumentClient(JsonDocumentOperations):
    def __init__(self, client: requests.Session):
        self._client = client

    def create_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
        upsert: bool = True,
        request_kwargs: Dict = {},
    ) -> JsonDocument:
        try:
            response = self._client.put(
                f"/projects/{project_id}/json/{collection_id}/{key}",
                data=json_document,
                params={"upsert": upsert},
                **request_kwargs,
            )
            handle_errors(response)
            return parse_raw_as(JsonDocument, response.text)
        except JSONDecodeError as ex:
            raise ClientValueError("The loaded JSON is invalid.") from ex

    def update_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
        request_kwargs: Dict = {},
    ) -> JsonDocument:
        try:
            response = self._client.patch(
                f"/projects/{project_id}/json/{collection_id}/{key}",
                data=json_document,
                **request_kwargs,
            )
            handle_errors(response)
            return parse_raw_as(JsonDocument, response.text)
        except JSONDecodeError as ex:
            raise ClientValueError("The loaded JSON is invalid.") from ex

    def list_json_documents(
        self,
        project_id: str,
        collection_id: str,
        filter: Optional[str] = None,
        keys: Optional[List[str]] = None,
        request_kwargs: Dict = {},
    ) -> List[JsonDocument]:
        response = self._client.get(
            f"/projects/{project_id}/json/{collection_id}",
            params={"filter": filter},
            # TODO: support filters
            **request_kwargs,
        )
        handle_errors(response)
        return parse_raw_as(List[JsonDocument], response.text)

    def get_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        request_kwargs: Dict = {},
    ) -> JsonDocument:
        response = self._client.get(
            f"/projects/{project_id}/json/{collection_id}/{key}", **request_kwargs
        )
        handle_errors(response)
        return parse_raw_as(JsonDocument, response.text)

    def delete_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self._client.delete(
            f"/projects/{project_id}/json/{collection_id}/{key}", **request_kwargs
        )
        handle_errors(response)

    def delete_json_collection(
        self,
        project_id: str,
        collection_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self._client.delete(
            f"/projects/{project_id}/json/{collection_id}", **request_kwargs
        )
        handle_errors(response)

    def delete_json_collections(
        self,
        project_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self._client.delete(f"/projects/{project_id}/json", **request_kwargs)
        handle_errors(response)



import re
from datetime import datetime, timezone
from typing import List

from loguru import logger

from contaxy import config
from contaxy.operations import AuthOperations, JsonDocumentOperations, ProjectOperations
from contaxy.operations.components import ComponentOperations
from contaxy.schema.auth import USERS_KIND, AccessLevel, User
from contaxy.schema.exceptions import (
    ClientValueError,
    ResourceAlreadyExistsError,
    ResourceNotFoundError,
)
from contaxy.schema.project import (
    MAX_PROJECT_ID_LENGTH,
    PROJECTS_KIND,
    Project,
    ProjectCreation,
    ProjectInput,
)
from contaxy.utils import auth_utils, id_utils


class ProjectManager(ProjectOperations):
    _PROJECT_COLLECTION = "projects"
    _PROJECT_NAME_TO_ID_REGEX = re.compile(r"^projects/([^/:\s]+)$")
    _MAX_ID_COUNTER = 9999

    def __init__(
        self,
        component_manager: ComponentOperations,
    ):
        """Initializes the project manager.

        Args:
            component_manager: Instance of the component manager that grants access to the other managers.
        """
        self._global_state = component_manager.global_state
        self._request_state = component_manager.request_state
        self._component_manager = component_manager

    @property
    def _json_db_manager(self) -> JsonDocumentOperations:
        return self._component_manager.get_json_db_manager()

    @property
    def _auth_manager(self) -> AuthOperations:
        return self._component_manager.get_auth_manager()

    def list_projects(self) -> List[Project]:
        authorized_user = None
        if self._request_state.authorized_access:
            authorized_user = self._request_state.authorized_access.authorized_subject
        projects: List[Project] = []
        if authorized_user:
            # Filter by the user
            resource_permissions = self._auth_manager.list_permissions(authorized_user)
            for permission in resource_permissions:
                resource_name, _ = auth_utils.parse_permission(permission)
                try:
                    project_id = id_utils.extract_project_id_from_resource_name(
                        resource_name
                    )
                    try:
                        project = self.get_project(project_id)
                        # If the project is a technical project, it is only returned for users with the `projects#admin` permission
                        if (
                            not project.technical_project
                            or auth_utils.construct_permission(
                                "projects", AccessLevel.ADMIN
                            )
                            in resource_permissions
                            or auth_utils.construct_permission("*", AccessLevel.ADMIN)
                            in resource_permissions
                        ):
                            projects.append(project)
                    except ResourceNotFoundError:
                        # this should not happen
                        logger.info(
                            f"Project not found: {project_id} during list projects."
                        )
                        continue
                except ValueError:
                    # do nothing
                    continue

        else:
            # TODO: also return all projects if user has admin role
            # If called without authorized user -> return all projects
            for json_document in self._json_db_manager.list_json_documents(
                config.SYSTEM_INTERNAL_PROJECT, self._PROJECT_COLLECTION
            ):
                projects.append(Project.parse_raw(json_document.json_value))

        return projects

    def create_project(
        self, project_input: ProjectCreation, technical_project: bool = False
    ) -> Project:
        try:
            # Check if project exists
            self.get_project(project_input.id)
            raise ResourceAlreadyExistsError(
                f"The project ID {project_input.id} is already used. Please select another project ID."
            )
        except ResourceNotFoundError:
            # This is expected, the project should not exist
            pass

        authorized_user = None
        if self._request_state.authorized_access:
            authorized_user = self._request_state.authorized_access.authorized_subject

        creation_timestamp = datetime.now(timezone.utc)
        project = Project(
            **project_input.dict(exclude_unset=True),
            created_at=creation_timestamp,
            updated_at=creation_timestamp,
            created_by=authorized_user,
            updated_by=authorized_user,
            technical_project=technical_project,
        )

        created_document = self._json_db_manager.create_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=self._PROJECT_COLLECTION,
            key=project_input.id,
            json_document=project.json(),
        )

        created_project = Project.parse_raw(created_document.json_value)

        if authorized_user:
            # Add admin permission for the project to the authorized user
            assert created_project.id is not None
            self.add_project_member(
                created_project.id,
                id_utils.extract_user_id_from_resource_name(authorized_user),
                AccessLevel.ADMIN,
            )
        return created_project

    def get_project(self, project_id: str) -> Project:
        try:
            json_document = self._json_db_manager.get_json_document(
                project_id=config.SYSTEM_INTERNAL_PROJECT,
                collection_id=self._PROJECT_COLLECTION,
                key=project_id,
            )
        except ResourceNotFoundError as ex:
            raise ResourceNotFoundError(
                f"Project not found for ID: {project_id}"
            ) from ex

        return Project.parse_raw(json_document.json_value)

    def update_project(self, project_id: str, project_input: ProjectInput) -> Project:
        updated_project = Project.parse_raw(project_input.json(exclude_unset=True))
        updated_project.updated_at = datetime.now(timezone.utc)
        if self._request_state.authorized_access:
            updated_project.updated_by = (
                self._request_state.authorized_access.authorized_subject
            )

        updated_document = self._json_db_manager.update_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=self._PROJECT_COLLECTION,
            key=project_id,
            json_document=updated_project.json(exclude_unset=True),
        )
        return Project.parse_raw(updated_document.json_value)

    def suggest_project_id(self, display_name: str) -> str:
        for i in range(self._MAX_ID_COUNTER):
            id_prefix = ""
            if i > 0:
                id_prefix = "-" + str(i)

            project_id = id_utils.generate_readable_id(
                display_name, max_length=MAX_PROJECT_ID_LENGTH, suffix=id_prefix
            )

            try:
                self.get_project(project_id)
            except ResourceNotFoundError:
                # No project exists with this ID -> return
                return project_id
        raise ClientValueError(
            f"Could not select project ID for name: {display_name}. Please try with another name."
        )

    def delete_project(self, project_id: str) -> None:
        # TODO: what to do on project deletion
        project_members = self.list_project_members(project_id)
        for project_member in project_members:
            # Remove all project permissions from all users
            self.remove_project_member(project_id, project_member.id)
        self._json_db_manager.delete_json_document(
            config.SYSTEM_INTERNAL_PROJECT, self._PROJECT_COLLECTION, project_id
        )

    def list_project_members(self, project_id: str) -> List[User]:
        project_member_resource_names: List[str] = []
        admin_permission = auth_utils.construct_permission(
            f"{PROJECTS_KIND}/{project_id}",
            AccessLevel.ADMIN,
        )
        project_member_resource_names.extend(
            self._auth_manager.list_resources_with_permission(
                admin_permission, resource_name_prefix=USERS_KIND
            )
        )

        write_permission = auth_utils.construct_permission(
            f"{PROJECTS_KIND}/{project_id}",
            AccessLevel.WRITE,
        )
        project_member_resource_names.extend(
            self._auth_manager.list_resources_with_permission(
                write_permission, resource_name_prefix=USERS_KIND
            )
        )

        read_permission = auth_utils.construct_permission(
            f"{PROJECTS_KIND}/{project_id}",
            AccessLevel.READ,
        )
        project_member_resource_names.extend(
            self._auth_manager.list_resources_with_permission(
                read_permission, resource_name_prefix=USERS_KIND
            )
        )

        project_users: List[User] = []

        for resource_name in project_member_resource_names:
            try:
                user_id = id_utils.extract_user_id_from_resource_name(resource_name)
            except ValueError:
                logger.warning(
                    "Failed to extract user id from resource name: " + resource_name
                )
                continue
            try:
                project_users.append(self._auth_manager.get_user(user_id))
            except ResourceNotFoundError:
                logger.warning(
                    f"User with id {user_id} does not exist anymore but its permissions have not been removed from the DB!"
                )
                continue

        return project_users

    def add_project_member(
        self,
        project_id: str,
        user_id: str,
        access_level: AccessLevel,
    ) -> List[User]:

        try:
            # Check if user with the given ID exists
            self._auth_manager.get_user(user_id)
        except ResourceNotFoundError as ex:
            raise ResourceNotFoundError(f"User {user_id} does not exist.") from ex

        self._auth_manager.add_permission(
            f"{USERS_KIND}/{user_id}",
            auth_utils.construct_permission(
                f"{PROJECTS_KIND}/{project_id}",
                access_level,
            )
            # TODO how to set the resource name and kind?
        )
        return self.list_project_members(project_id)

    def _remove_project_member(self, project_id: str, user_id: str) -> None:
        self._auth_manager.remove_permission(
            f"{USERS_KIND}/{user_id}",
            auth_utils.construct_permission(
                f"{PROJECTS_KIND}/{project_id}", AccessLevel.ADMIN
            ),
            remove_sub_permissions=True,
        )

    def remove_project_member(self, project_id: str, user_id: str) -> List[User]:
        # Remove all permissions from the user that grant access to any part of the project
        self._remove_project_member(project_id, user_id)
        return self.list_project_members(project_id)



import json
import time
from datetime import datetime, timezone
from typing import Dict, List, Literal, Optional, Union

from starlette.responses import Response

from contaxy import config
from contaxy.config import settings
from contaxy.managers.deployment.docker import DockerDeploymentPlatform
from contaxy.managers.deployment.kubernetes import KubernetesDeploymentPlatform
from contaxy.managers.deployment.utils import (
    create_deployment_config,
    get_job_collection_id,
    get_service_collection_id,
    split_image_name_and_tag,
)
from contaxy.operations import (
    AuthOperations,
    DeploymentOperations,
    JsonDocumentOperations,
    SystemOperations,
)
from contaxy.operations.components import ComponentOperations
from contaxy.schema import (
    ClientValueError,
    Job,
    JobInput,
    ResourceAction,
    ResourceNotFoundError,
    ServerBaseError,
    Service,
    ServiceInput,
)
from contaxy.schema.deployment import DeploymentStatus, DeploymentType, ServiceUpdate
from contaxy.schema.shared import ResourceActionExecution

ACTION_DELIMITER = "-"
ACTION_ACCESS = "access"
ACTION_START = "start"
ACTION_STOP = "stop"
ACTION_RESTART = "restart"


class DeploymentManager(DeploymentOperations):
    def __init__(
        self,
        deployment_platform: Union[
            DockerDeploymentPlatform, KubernetesDeploymentPlatform
        ],
        component_manager: ComponentOperations,
    ):
        self._global_state = component_manager.global_state
        self._request_state = component_manager.request_state
        self.deployment_platform = deployment_platform
        self._component_manager = component_manager

    @property
    def _json_db_manager(self) -> JsonDocumentOperations:
        return self._component_manager.get_json_db_manager()

    @property
    def _system_manager(self) -> SystemOperations:
        return self._component_manager.get_system_manager()

    @property
    def _auth_manager(self) -> AuthOperations:
        return self._component_manager.get_auth_manager()

    def deploy_service(
        self,
        project_id: str,
        service_input: ServiceInput,
        action_id: Optional[str] = None,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
        wait: bool = False,
    ) -> Service:
        # Create service in DB
        db_service: Service = create_deployment_config(
            project_id=project_id,
            deployment_input=service_input,
            deployment_type=deployment_type,
            authorized_subject=self._request_state.authorized_subject,
            system_manager=self._system_manager,
            auth_manager=self._auth_manager,
            deployment_class=Service,
        )
        self._create_service_db_document(db_service, project_id)

        # Start service container
        if not service_input.is_stopped:
            deployed_service = self._deploy_service(
                project_id, db_service, action_id, wait
            )
            db_service.status = deployed_service.status
            db_service.internal_id = deployed_service.internal_id
        else:
            db_service.status = DeploymentStatus.STOPPED
            db_service.internal_id = None
        return db_service

    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> List[Service]:
        service_docs = self._json_db_manager.list_json_documents(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_service_collection_id(project_id),
        )
        # Create lookup for all running services
        deployed_service_lookup: Dict[str, Service] = {
            service.id: service
            for service in self.deployment_platform.list_services(
                project_id, deployment_type
            )
            if service.id is not None
        }
        services = []
        # Go through all services in the DB and update their status and internal id
        for service_doc in service_docs:
            db_service = Service.parse_raw(service_doc.json_value)
            if db_service.deployment_type != deployment_type:
                continue
            deployed_service = deployed_service_lookup.pop(db_service.id, None)
            if deployed_service is None:
                db_service.status = DeploymentStatus.STOPPED
                db_service.internal_id = None
            else:
                db_service.status = deployed_service.status
                db_service.internal_id = deployed_service.internal_id
            services.append(db_service)
        # Add services to DB that were not added via the contaxy API
        for deployed_service in deployed_service_lookup.values():
            self._create_service_db_document(deployed_service, project_id)
            services.append(deployed_service)

        return services

    def _create_service_db_document(self, service: Service, project_id: str) -> None:
        self._json_db_manager.create_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_service_collection_id(project_id),
            key=service.id,
            json_document=service.json(exclude={"status", "internal_id"}),
            upsert=False,
        )

    def get_service_metadata(self, project_id: str, service_id: str) -> Service:
        db_service = self._get_service_from_db(project_id, service_id)
        try:
            deployed_service = self.deployment_platform.get_service_metadata(
                project_id, db_service.id
            )
            db_service.status = deployed_service.status
            db_service.internal_id = deployed_service.internal_id
        except ResourceNotFoundError:
            db_service.status = DeploymentStatus.STOPPED
            db_service.internal_id = None

        return db_service

    def update_service(
        self, project_id: str, service_id: str, service: ServiceUpdate
    ) -> Service:
        service_dict = service.dict(exclude_unset=True)
        if "display_name" in service_dict:
            raise ClientValueError("Display name of service cannot be updated!")
        if "container_image" in service_dict:
            image_name, image_tag = split_image_name_and_tag(service.container_image)
            self._system_manager.check_allowed_image(image_name, image_tag)
        service_doc = self._json_db_manager.update_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_service_collection_id(project_id),
            key=service_id,
            json_document=service.json(exclude_unset=True),
        )
        db_service = Service.parse_raw(service_doc.json_value)
        try:
            deployed_service = self._execute_restart_service_action(
                project_id, service_id
            )
            db_service.status = deployed_service.status
            db_service.internal_id = deployed_service.internal_id
        except ClientValueError:
            db_service.status = DeploymentStatus.STOPPED
            db_service.internal_id = None

        return db_service

    def update_service_access(self, project_id: str, service_id: str) -> None:
        user = self._request_state.authorized_subject
        self._json_db_manager.update_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_service_collection_id(project_id),
            key=service_id,
            json_document=json.dumps(
                {
                    "last_access_time": str(datetime.now(timezone.utc)),
                    "last_access_user": user,
                }
            ),
        )

    def delete_service(
        self, project_id: str, service_id: str, delete_volumes: bool = False
    ) -> None:
        db_service = self._get_service_from_db(project_id, service_id)

        try:
            self.deployment_platform.delete_service(
                project_id=project_id,
                service_id=db_service.id,
                delete_volumes=delete_volumes,
            )
        except ResourceNotFoundError:
            # Workspace is already stopped and service does not exist
            pass

        self._json_db_manager.delete_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_service_collection_id(project_id),
            key=db_service.id,
        )

    def delete_services(self, project_id: str) -> None:
        self.deployment_platform.delete_services(project_id)
        self._json_db_manager.delete_json_collection(
            config.SYSTEM_INTERNAL_PROJECT, get_service_collection_id(project_id)
        )

    def _get_service_from_db(self, project_id: str, service_id: str) -> Service:
        try:
            service_doc = self._json_db_manager.get_json_document(
                project_id=config.SYSTEM_INTERNAL_PROJECT,
                collection_id=get_service_collection_id(project_id),
                key=service_id,
            )
        except ResourceNotFoundError:
            raise ResourceNotFoundError(
                f"The service with id {service_id} could not "
                f"be found in project {project_id}!"
            )
        db_service = Service.parse_raw(service_doc.json_value)
        return db_service

    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int],
        since: Optional[datetime],
    ) -> str:
        return self.deployment_platform.get_service_logs(
            project_id, service_id, lines, since
        )

    def list_jobs(self, project_id: str) -> List[Job]:
        job_docs = self._json_db_manager.list_json_documents(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_job_collection_id(project_id),
        )
        deployed_job_lookup: Dict[Optional[str], Job] = {
            job.id: job for job in self.deployment_platform.list_jobs(project_id)
        }
        jobs = []
        for job_doc in job_docs:
            db_job = Job.parse_raw(job_doc.json_value)
            deployed_job = deployed_job_lookup.get(db_job.id)
            if deployed_job is None:
                db_job.status = DeploymentStatus.STOPPED
                db_job.internal_id = None
            else:
                db_job.status = deployed_job.status
                db_job.internal_id = deployed_job.internal_id
            jobs.append(db_job)

        return jobs

    def deploy_job(
        self,
        project_id: str,
        job_input: JobInput,
        action_id: Optional[str] = None,
        wait: bool = False,
    ) -> Job:
        # Create job in DB
        db_job: Job = create_deployment_config(
            project_id=project_id,
            deployment_input=job_input,
            deployment_type=DeploymentType.JOB,
            authorized_subject=self._request_state.authorized_subject,
            system_manager=self._system_manager,
            auth_manager=self._auth_manager,
            deployment_class=Job,
        )
        self._json_db_manager.create_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_job_collection_id(project_id),
            key=db_job.id,
            json_document=db_job.json(),
            upsert=False,
        )
        # Start job container
        deployed_job = self.deployment_platform.deploy_job(
            project_id, db_job, action_id, wait
        )
        db_job.status = deployed_job.status
        db_job.internal_id = deployed_job.internal_id
        return db_job

    def get_job_metadata(self, project_id: str, job_id: str) -> Job:
        db_job = self._get_job_from_db(project_id, job_id)
        try:
            deployed_job = self.deployment_platform.get_job_metadata(
                project_id, db_job.id
            )
            db_job.status = deployed_job.status
            db_job.internal_id = deployed_job.internal_id
        except ResourceNotFoundError:
            db_job.status = DeploymentStatus.STOPPED
            db_job.internal_id = None

        return db_job

    def delete_job(self, project_id: str, job_id: str) -> None:
        db_job = self._get_job_from_db(project_id, job_id)

        try:
            self.deployment_platform.delete_job(
                project_id=project_id,
                job_id=db_job.id,
            )
        except ResourceNotFoundError:
            pass

        self._json_db_manager.delete_json_document(
            project_id=config.SYSTEM_INTERNAL_PROJECT,
            collection_id=get_job_collection_id(project_id),
            key=db_job.id,
        )

    def delete_jobs(self, project_id: str) -> None:
        self.deployment_platform.delete_jobs(project_id)
        self._json_db_manager.delete_json_collection(
            config.SYSTEM_INTERNAL_PROJECT, get_job_collection_id(project_id)
        )

    def _get_job_from_db(self, project_id: str, job_id: str) -> Job:
        try:
            job_doc = self._json_db_manager.get_json_document(
                project_id=config.SYSTEM_INTERNAL_PROJECT,
                collection_id=get_job_collection_id(project_id),
                key=job_id,
            )
        except ResourceNotFoundError:
            raise ResourceNotFoundError(
                f"The job with id {job_id} could not "
                f"be found in project {project_id}!"
            )
        db_job = Job.parse_raw(job_doc.json_value)
        return db_job

    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        return self.deployment_platform.get_job_logs(project_id, job_id, lines, since)

    def execute_service_action(
        self,
        project_id: str,
        service_id: str,
        action_id: str,
        action_execution: ResourceActionExecution = ResourceActionExecution(),
    ) -> Response:
        if action_id == ACTION_START:
            self._execute_start_service_action(project_id, service_id)
        elif action_id == ACTION_STOP:
            self._execute_stop_service_action(project_id, service_id)
        elif action_id == ACTION_RESTART:
            self._execute_restart_service_action(project_id, service_id)
        else:
            return Response(
                content=f"No implementation for action id '{action_id}'",
                status_code=501,
            )
        return Response(
            content=f"Action {action_id} successfully executed.", status_code=200
        )

    def _execute_start_service_action(self, project_id: str, service_id: str) -> None:
        service = self.get_service_metadata(project_id, service_id)
        if service.status != DeploymentStatus.STOPPED:
            raise ClientValueError(
                f"Action {ACTION_START} on service {service_id} can only be performed "
                f"when service is stopped but status is {service.status}!"
            )
        self._deploy_service(project_id, service, wait=True)

    def _deploy_service(
        self,
        project_id: str,
        service: Service,
        action_id: Optional[str] = None,
        wait: bool = False,
    ) -> Service:
        # Make sure service access time is updated before starting the container
        # to avoid it being considered idle immediately due to an only access time
        self.update_service_access(project_id, service.id)
        return self.deployment_platform.deploy_service(
            project_id, service, action_id=action_id, wait=wait
        )

    def _execute_stop_service_action(
        self,
        project_id: str,
        service_id: str,
    ) -> None:
        service = self.get_service_metadata(project_id, service_id)
        if service.status == DeploymentStatus.STOPPED:
            raise ClientValueError(
                f"Action {ACTION_STOP} on service {service_id} can only be performed "
                f"when service is not stopped already!"
            )
        self.deployment_platform.delete_service(
            project_id, service_id, delete_volumes=service.clear_volume_on_stop
        )

    def _execute_restart_service_action(
        self, project_id: str, service_id: str
    ) -> Service:
        service = self.get_service_metadata(project_id, service_id)
        if service.status == DeploymentStatus.STOPPED:
            raise ClientValueError(
                f"Action {ACTION_RESTART} on service {service_id} can only be performed "
                f"when service is not stopped already!"
            )
        return self._restart_service(project_id, service)

    def _restart_service(self, project_id: str, service: Service) -> Service:
        self.deployment_platform.delete_service(
            project_id, service.id, delete_volumes=False
        )
        self._wait_for_deployment_deletion(project_id, service.id)
        return self._deploy_service(project_id, service, wait=True)

    def _wait_for_deployment_deletion(
        self, project_id: str, service_id: str, timeout_seconds: int = 60
    ) -> None:
        try:
            start_time = time.time()
            while time.time() - start_time < timeout_seconds:
                self.deployment_platform.get_service_metadata(project_id, service_id)
                time.sleep(3)
            raise ServerBaseError(
                f"Action {ACTION_RESTART} on service {service_id} failed as service "
                f"did not stop after waiting {timeout_seconds} seconds."
            )
        except ResourceNotFoundError:
            # Service was successfully deleted
            pass

    def execute_job_action(
        self,
        project_id: str,
        job_id: str,
        action_id: str,
        action_execution: ResourceActionExecution = ResourceActionExecution(),
    ) -> Response:
        # 501: not implemented
        return Response(status_code=501)

    def list_service_actions(
        self, project_id: str, service_id: str
    ) -> List[ResourceAction]:
        service_metadata = self.get_service_metadata(
            project_id=project_id, service_id=service_id
        )

        resource_actions: List[ResourceAction] = []
        if service_metadata.status == DeploymentStatus.STOPPED:
            resource_actions.append(
                ResourceAction(action_id=ACTION_START, display_name="Start Service")
            )
        else:
            resource_actions.append(
                ResourceAction(action_id=ACTION_STOP, display_name="Stop Service")
            )
            resource_actions.append(
                ResourceAction(action_id=ACTION_RESTART, display_name="Restart Service")
            )
            resource_actions += self._get_service_access_actions(
                project_id, service_id, service_metadata
            )

        return resource_actions

    @staticmethod
    def _get_service_access_actions(
        project_id: str, service_id: str, service_metadata: Service
    ) -> List[ResourceAction]:
        if not service_metadata.endpoints:
            return []
        access_actions = []
        for endpoint in service_metadata.endpoints:
            if len(service_metadata.endpoints) > 1:
                display_name = f"Access Service on Endpoint: {endpoint}"
            else:
                display_name = "Access Service"
            access_actions.append(
                ResourceAction(
                    action_id=f"{ACTION_ACCESS}{ACTION_DELIMITER}{endpoint.replace('/', '')}",
                    display_name=display_name,
                    instructions=[
                        {
                            "type": "new-tab",
                            "url": f"{settings.CONTAXY_BASE_URL}/projects/{project_id}/services/{service_id}/access/{endpoint}",
                        }
                    ],
                )
            )
        return access_actions

    def list_deploy_service_actions(
        self, project_id: str, service: ServiceInput
    ) -> List[ResourceAction]:
        return self.deployment_platform.list_deploy_service_actions(project_id, service)

    def list_deploy_job_actions(
        self, project_id: str, job: JobInput
    ) -> List[ResourceAction]:
        return self.deployment_platform.list_deploy_job_actions(project_id, job)

    def list_job_actions(
        self,
        project_id: str,
        job_id: str,
    ) -> List[ResourceAction]:
        return []

    def suggest_service_config(
        self,
        project_id: str,
        container_image: str,
    ) -> ServiceInput:
        # TODO: add sensible logic here
        return ServiceInput(container_image=container_image)

    def suggest_job_config(self, project_id: str, container_image: str) -> JobInput:
        # TODO: add sensible logic here
        return JobInput(container_image=container_image)



import requests
from fastapi import HTTPException, status
from requests.models import Response

from contaxy.schema import UnauthenticatedError
from contaxy.schema.exceptions import (
    ClientValueError,
    PermissionDeniedError,
    ProblemDetails,
    ResourceAlreadyExistsError,
    ResourceNotFoundError,
)


class BaseUrlSession(requests.Session):
    def __init__(self, base_url=None, *args, **kwargs):  # type: ignore
        super(BaseUrlSession, self).__init__(*args, **kwargs)
        self.base_url = base_url

    def request(self, method, url, *args, **kwargs):  # type: ignore
        url = self.base_url + url
        return super(BaseUrlSession, self).request(method, url, *args, **kwargs)


def handle_errors(response: Response) -> None:
    if response.status_code < 400 or response.status_code >= 600:
        return

    message = None
    try:
        error = ProblemDetails.parse_raw(response.text)
        message = error.message
    except Exception:
        try:
            # Use full body
            message = response.text
        except Exception:
            pass

    if response.status_code == status.HTTP_401_UNAUTHORIZED:
        raise UnauthenticatedError(message)

    if response.status_code == status.HTTP_403_FORBIDDEN:
        raise PermissionDeniedError(message)

    if response.status_code == status.HTTP_404_NOT_FOUND:
        raise ResourceNotFoundError(message)

    if response.status_code == status.HTTP_409_CONFLICT:
        raise ResourceAlreadyExistsError(message)

    # TODO: already used
    # if response.status_code == status.HTTP_409_CONFLICT:
    #    raise ResourceUpdateFailedError(message)

    if response.status_code == status.HTTP_400_BAD_REQUEST:
        raise ClientValueError(message)

    # If different error, raise generic http exception
    # This should not happen
    raise HTTPException(status_code=response.status_code, detail=response.text)



import io
from random import randint
from typing import List, Optional

from faker import Faker
from pydantic import EmailStr

from contaxy.operations import AuthOperations, FileOperations, ProjectOperations
from contaxy.operations.components import ComponentOperations
from contaxy.operations.seed import SeedOperations
from contaxy.schema import File, Project, User, UserRegistration
from contaxy.schema.project import ProjectCreation
from contaxy.utils import auth_utils
from contaxy.utils.file_utils import FileStreamWrapper

FAKER = Faker()
Faker.seed(0)

ERROR_NO_PROJECT_MANAGER = "Seeder needs to be initialized with a project manager"


class SeedManager(SeedOperations):
    def __init__(
        self,
        component_manager: ComponentOperations,
    ):
        self.global_state = component_manager.global_state
        self.request_state = component_manager.request_state
        self._component_manager = component_manager

    @property
    def file_manager(self) -> FileOperations:
        return self._component_manager.get_file_manager()

    @property
    def auth_manager(self) -> AuthOperations:
        return self._component_manager.get_auth_manager()

    @property
    def project_manager(self) -> ProjectOperations:
        return self._component_manager.get_project_manager()

    def create_user(
        self,
        user_input: UserRegistration = UserRegistration(
            username="Foo", email=EmailStr("foo@bar.com"), password="Foobar"
        ),
    ) -> User:
        if not self.auth_manager:
            raise RuntimeError("Seeder needs to be initialized with an auth manager")
        if not self.project_manager:
            raise RuntimeError(ERROR_NO_PROJECT_MANAGER)

        user = auth_utils.create_and_setup_user(
            user_input=user_input,
            auth_manager=self.auth_manager,
            project_manager=self.project_manager,
        )
        return user

    def create_users(self, amount: int) -> List[User]:
        users = []
        for _ in range(amount):
            user = self.create_user(
                UserRegistration(
                    username=FAKER.user_name(),
                    email=FAKER.email(),
                    password=FAKER.password(),
                )
            )
            users.append(user)

        return users

    def create_project(
        self,
        project_input: ProjectCreation = ProjectCreation(
            id="my-test-project", display_name="My Test Project!"
        ),
    ) -> Project:
        if not self.project_manager:
            raise RuntimeError(ERROR_NO_PROJECT_MANAGER)
        return self.project_manager.create_project(project_input=project_input)

    def create_projects(self, amount: int) -> List[Project]:
        if not self.project_manager:
            raise RuntimeError(ERROR_NO_PROJECT_MANAGER)
        projects = []
        for _ in range(amount):
            project = self.create_project(
                ProjectCreation(id=FAKER.ean(), display_name=FAKER.word())
            )
            projects.append(project)

        return projects

    def create_file(
        self,
        project_id: str,
        file_key: str = "my-test-file",
        max_number_chars: int = 200,
    ) -> File:
        if not self.file_manager:
            raise RuntimeError("Seeder needs to be initialized with a file manager")
        file_stream = FileStreamWrapper(
            io.BytesIO(FAKER.text(max_nb_chars=max_number_chars).encode("UTF-8"))
        )
        return self.file_manager.upload_file(
            project_id=project_id, file_key=file_key, file_stream=file_stream  # type: ignore
        )

    def create_files(
        self,
        project_id: str,
        number_of_files: int,
        prefix: Optional[str] = "my-test-file",
        max_number_chars: int = 200,
    ) -> List[File]:
        return [
            self.create_file(
                project_id,
                f"{prefix}-{randint(1, 10000)}",
                max_number_chars,
            )
            for index in range(number_of_files)
        ]

    def create_file_stream(
        self,
        max_number_chars: int = 200,
    ) -> FileStreamWrapper:
        return FileStreamWrapper(
            io.BytesIO(FAKER.text(max_nb_chars=max_number_chars).encode("UTF-8"))
        )



import json
import os
from typing import Dict, Iterator, List, Optional, Tuple

from loguru import logger
from minio import Minio
from minio.datatypes import Object as MinioObject
from minio.deleteobjects import DeleteObject
from minio.error import S3Error
from starlette.responses import Response

from contaxy.operations import FileOperations
from contaxy.operations.components import ComponentOperations
from contaxy.operations.json_db import JsonDocumentOperations
from contaxy.schema import File, FileInput, ResourceAction
from contaxy.schema.exceptions import (
    ClientValueError,
    ResourceNotFoundError,
    ServerBaseError,
)
from contaxy.schema.file import FileStream
from contaxy.schema.json_db import JsonDocument
from contaxy.utils.file_utils import generate_file_id
from contaxy.utils.minio_utils import (
    create_bucket,
    create_minio_client,
    delete_bucket,
    get_bucket_name,
)


class MinioFileManager(FileOperations):

    DOC_COLLECTION_NAME = "s3_files"
    METADATA_FIELD_SET = {
        "key",
        "external_id",
        "md5_hash",
        "description",
        "content_type",
        "metadata",
        "disabled",
        "icon",
        "display_name",
        "tags",
    }

    def __init__(
        self,
        component_manager: ComponentOperations,
    ):
        """Initializes the Minio File Manager.

        Args:
            component_manager: Instance of the component manager that grants access to the other managers.
        """
        self._global_state = component_manager.global_state
        self._request_state = component_manager.request_state
        self._component_manager = component_manager
        self.client = self._create_client()
        self.sys_namespace = self._global_state.settings.SYSTEM_NAMESPACE

    @property
    def _json_db_manager(self) -> JsonDocumentOperations:
        return self._component_manager.get_json_db_manager()

    def list_files(
        self,
        project_id: str,
        recursive: bool = True,
        include_versions: bool = False,
        prefix: Optional[str] = None,
    ) -> List[File]:
        """List files.

        Args:
            project_id (str): Project ID associated with the file.
            recursive (bool, optional): List recursively as directory structure emulation. Defaults to True.
            include_versions (bool, optional): Flag to control whether include object versions. Defaults to False.
            prefix (Optional[str], optional): File key starts with prefix. Defaults to None.

        Returns:
            List[File]: List of file metadata objects.
        """

        # TODO: Test case when object is folder
        logger.debug(
            f"list_files (`project_id`: {project_id}, `recursive`: {recursive}, `include_versions`: {include_versions}, `prefix`: {prefix})."
        )

        bucket_name = get_bucket_name(project_id, self.sys_namespace)

        file_data, document_keys = self._load_file_data_from_s3(
            bucket_name, prefix, recursive, include_versions
        )

        file_data, docs_not_in_db = self._enrich_data_from_db(
            project_id, file_data, document_keys
        )

        # Usually each file should have a corresponding db entry with some meta data.
        # This might not be the case, if someone manually added a file to S3.
        # TODO: Consider conflict handling
        for key, json_doc in docs_not_in_db:
            self._json_db_manager.create_json_document(
                project_id, self.DOC_COLLECTION_NAME, key, json_doc
            )

        return file_data

    def get_file_metadata(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
    ) -> File:
        """Get file metadata of a single file.

        If no version is provided then the latest version will be returned.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.

        Raises:
            ResourceNotFoundError: If no file is found.

        Returns:
            File: The file metadata object.
        """
        logger.debug(
            f"get_file_metadata (`project_id`: {project_id}, `file_key`: {file_key}, `version`: {version})"
        )

        files_to_prefix = self.list_files(
            project_id, prefix=file_key, include_versions=True if version else False
        )

        file_versions = filter(lambda file: file.key == file_key, files_to_prefix)

        for file in file_versions:
            if not version or (file.version == version):
                return file

        msg = f"File not found (key: {file_key}"
        msg = f"{msg}, version {version})." if version else f"{msg})."
        raise ResourceNotFoundError(msg)

    def update_file_metadata(
        self,
        file: FileInput,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
    ) -> File:
        """Update the file metadata.

        If no version is provided then the latest version will be returned. Moreover, additional custom metadata provied via `file.metadata` will executed with json merge patch strategy in case a specific version is updated. But if a new version was created it is treated as a new file and hence the metadata from older versions will not be considered. If a version should inherit the metadata from a previous version, then this data needs to be set explictly.

        Args:
            file (FileInput): The file metadata object. All unset attributes or None values will be ignored.
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.

        Raises:
            ClientValueError: If the provided keys do not match.

        Returns:
            File: The updated file metadata object.
        """
        logger.debug(
            f"update_file_metadata (`project_id`: {project_id}, `file_key`: {file_key}, `version`: {version})"
        )

        if file.key != file_key:
            raise ClientValueError(
                f"File keys do not match (file_key: {file_key}, file.key {file.key})"
            )

        try:
            s3_object = self.client.stat_object(
                get_bucket_name(
                    project_id, self._global_state.settings.SYSTEM_NAMESPACE
                ),
                file_key,
                version_id=version,
            )
        except S3Error as err:
            if err.code == "NoSuchKey":
                raise ResourceNotFoundError(
                    f"Invalid file {file_key} (version: {version}."
                )
            raise err

        data = file.dict(
            include=self.METADATA_FIELD_SET, exclude_none=True, exclude_unset=True
        )
        # TODO: Fix metadata creation
        json_value = json.dumps(data)

        doc_key = generate_file_id(file_key, s3_object.version_id)

        try:
            self._json_db_manager.update_json_document(
                project_id, self.DOC_COLLECTION_NAME, doc_key, json_value
            )
        except ResourceNotFoundError:
            # Lazily create and update a metadata entry in the DB, since the resource might be uploaded externally
            self._create_file_metadata_json_document(
                project_id, file_key, s3_object.version_id
            )
            self._json_db_manager.update_json_document(
                project_id, self.DOC_COLLECTION_NAME, doc_key, json_value
            )

        return self.get_file_metadata(project_id, file_key, s3_object.version_id)

    def upload_file(
        self,
        project_id: str,
        file_key: str,
        file_stream: FileStream,
        content_type: str = "application/octet-stream",
    ) -> File:
        """Upload a file.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            file_stream (FileStream): The actual file stream object.
            content_type (str, optional): The mime-type of the file. Defaults to "application/octet-stream".

        Raises:
            ServerBaseError: If the upload failed.

        Returns:
            File: The file metadata object of the uploaded file.
        """

        logger.debug(
            f"upload_file (`project_id`: {project_id}, `file_key`: {file_key})"
        )

        # Todo: Handle further metadata: creation, disabled, tags, icon, userdata, extension_id
        bucket_name = get_bucket_name(project_id, self.sys_namespace)

        if not self.client.bucket_exists(bucket_name):
            # This can not be done lazily because put_object still consumes the stream although the bucket does not exists yet
            logger.debug(f"No bucket existing for project {project_id}.")
            create_bucket(self.client, bucket_name)

        try:
            result = self.client.put_object(
                bucket_name,
                file_key,
                file_stream,
                -1,
                content_type,
                part_size=10 * 1024 * 1024,
            )
        except S3Error as err:
            raise ServerBaseError(
                f"The file {file_key} could not be uploaded ({err.code})."
            )

        if hasattr(file_stream, "hash"):
            file_hash: Optional[str] = file_stream.hash
        else:
            file_hash = None
            logger.error(
                "The provided stream object does not provide a hash property. No hash will be available."
            )

        # ? Get the data from the last version if existing?
        self._create_file_metadata_json_document(
            project_id, file_key, result.version_id, file_hash
        )

        # This is necessary in order to have the available versions metadata set
        files_to_prefix = self.list_files(
            project_id, recursive=False, include_versions=False, prefix=file_key
        )
        file_versions = filter(lambda file: file.key == file_key, files_to_prefix)
        return file_versions.__next__()

    def download_file(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
    ) -> Iterator[bytes]:
        """Download a file.

        Either the latest version will be returned or the specified one.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.

        Raises:
            ResourceNotFoundError: If file does not exist.

        Yields:
            Iterator[bytes]: [description]
        """
        logger.debug(
            f"download_file (`project_id`: {project_id}, `file_key`: {file_key}, `version`: {version})"
        )
        bucket_name = get_bucket_name(project_id, self.sys_namespace)
        try:
            response = self.client.get_object(bucket_name, file_key, version_id=version)
        except S3Error as err:
            if err.code == "NoSuchBucket":
                logger.debug(f"Download failed - Bucket {bucket_name} does not exist.")
                raise ResourceNotFoundError(f"The project {project_id} is invalid.")
            elif err.code == "NoSuchKey":
                logger.debug(
                    f"Download failed - Invalid file {file_key} (version: {version})."
                )
                raise ResourceNotFoundError(
                    f"Invalid file {file_key} (version: {version}."
                )

        return response.stream()

    def delete_file(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        keep_latest_version: bool = False,
    ) -> None:
        """Delete a file.

        If a specific file `version` is provided, only this one will be deleted. If no `version` is provided and `keep_latest_version` is True, all but the latest version will be deleted. Otherwise, all existing versions will be removed.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.
            keep_latest_version (bool, optional): [description]. Defaults to False.
        """

        logger.debug(
            f"delete_file (`project_id`: {project_id}, `file_key`: {file_key}, `version`: {version}, `keep_latest_version`: {keep_latest_version})"
        )

        if keep_latest_version or version is None:
            self._delete_file_versions(project_id, file_key, keep_latest_version)
            return

        bucket_name = get_bucket_name(project_id, self.sys_namespace)

        try:
            # Does not work without version
            self.client.remove_object(bucket_name, file_key, version_id=version)
        except S3Error as err:
            if err.code == "NoSuchKey":
                logger.debug(
                    f"Deletion failed - No file {file_key} exists in bucket {bucket_name}."
                )
            elif err.code == "NoSuchBucket":
                logger.debug(f"Deletion failed - Bucket {bucket_name} does not exist.")
            raise err

        try:
            self._json_db_manager.delete_json_document(
                project_id,
                self.DOC_COLLECTION_NAME,
                generate_file_id(file_key, version),
            )
        except ResourceNotFoundError:
            logger.warning(
                f"No file metadata Json doc found for file_key: {file_key}, version: {version}"
            )

    def delete_files(
        self,
        project_id: str,
    ) -> None:
        """Delete all files and storage resources related to a project.

        Args:
            project_id (str): Project ID associated with the files.
        """
        delete_bucket(
            self.client,
            get_bucket_name(project_id, self._global_state.settings.SYSTEM_NAMESPACE),
            force=True,
        )
        self._json_db_manager.delete_json_collection(
            project_id, self.DOC_COLLECTION_NAME
        )

    def list_file_actions(
        self, project_id: str, file_key: str, version: Optional[str] = None
    ) -> List[ResourceAction]:
        return []

    def execute_file_action(
        self,
        project_id: str,
        file_key: str,
        action_id: str,
        version: Optional[str] = None,
    ) -> Response:
        # TODO
        pass

    def _create_client(self) -> Minio:
        settings = self._global_state.settings
        state_namespace = self._request_state[MinioFileManager]
        if not state_namespace.client:
            state_namespace.client = create_minio_client(self._global_state.settings)
            logger.info(
                f"Minio client created (endpoint: {settings.S3_ENDPOINT}, region: {settings.S3_REGION}, secure: {settings.S3_SECURE})"
            )
        return state_namespace.client

    def _create_file_metadata_json_document(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        md5_hash: Optional[str] = None,
    ) -> JsonDocument:
        s3_object = self.client.stat_object(
            get_bucket_name(project_id, self._global_state.settings.SYSTEM_NAMESPACE),
            file_key,
            version_id=version,
        )
        meta_file = self._map_s3_object_to_file_model(s3_object)
        if md5_hash:
            meta_file.md5_hash = md5_hash
        metadata_json = self._map_file_obj_to_json_document(meta_file)
        return self._json_db_manager.create_json_document(
            project_id,
            self.DOC_COLLECTION_NAME,
            meta_file.id,  # type: ignore
            metadata_json,
        )

    def _map_s3_object_to_file_model(self, object: MinioObject) -> File:
        # Todo: Check if directory can be given and what the implications are. Do we want to list such? (see object.is_dir param)
        file_extension = os.path.splitext(object.object_name)[1]
        display_name = os.path.basename(object.object_name)
        data = {
            "id": generate_file_id(object.object_name, object.version_id),
            "external_id": object.object_name,
            "key": object.object_name,
            "display_name": display_name,
            "updated_at": object.last_modified,
            "file_extension": file_extension,
            "file_size": object.size,
            "etag": object.etag,
            "latest_version": object.is_latest,
            "version": object.version_id,
        }

        # ! Minio seems to not set the content type, only when reading single object via stat_object, hence we persist it in the db
        if object.content_type:
            data.update({"content_type": object.content_type})
        return File(**data)

    def _map_file_obj_to_json_document(self, file: File) -> str:
        # Todo: Handle creation metadata. Problem: Each version is actually a new object and gets a dedicated db entry. Use list_objects or direct lookup in the DB?
        json_data = json.dumps(
            file.dict(include=self.METADATA_FIELD_SET, exclude_none=True),
            default=str,
        )
        return json_data

    def _load_file_data_from_s3(
        self,
        bucket_name: str,
        prefix: Optional[str] = None,
        recursive: Optional[bool] = None,
        include_version: Optional[bool] = None,
    ) -> Tuple[List[File], List[str]]:
        # TODO: Currently the available version field is always set and thus all versions are internally read everytime. Decide what is the desired behavior here.

        file_data: List[File] = []
        db_keys: List[str] = []
        file_versions: Dict[str, List[str]] = {}

        try:
            objects: List[MinioObject] = self.client.list_objects(
                bucket_name,
                prefix,
                recursive,
                include_version=True,
                include_user_meta=True,
            )

            for obj in objects:
                file = self._map_s3_object_to_file_model(obj)
                if include_version or file.latest_version:
                    file_data.append(file)
                    # ? The DB Keys might be added always in order to improve syncing between S3 and DB. But this would imply to read all versions from the DB everytime, even when versions arent requested by the user.
                    db_keys.append(str(file.id))

                versions = file_versions.get(file.key, [])
                file_versions.update({file.key: versions + [str(file.version)]})

            if not file_data:
                return [], []

        except S3Error as err:
            if err.code == "NoSuchBucket":
                logger.debug(f"No files to list - Bucket {bucket_name} does not exist.")
                return [], []

        for file in file_data:
            file.available_versions = file_versions.get(file.key)

        return file_data, db_keys

    def _enrich_data_from_db(
        self, project_id: str, file_data: List[File], document_keys: List[str]
    ) -> Tuple[List[File], List[Tuple[str, str]]]:

        json_docs = self._json_db_manager.list_json_documents(
            project_id, self.DOC_COLLECTION_NAME, keys=document_keys
        )

        # Create dict for efficient lookups
        doc_dict: Dict[str, JsonDocument] = {}
        for doc in json_docs:
            doc_dict.update({doc.key: doc})

        docs_not_in_db: List[Tuple[str, str]] = []
        for file in file_data:
            doc_of_file = doc_dict.get(str(file.id))
            if doc_of_file:
                file = self._add_data_from_doc_to_file(file, doc_of_file)
            else:
                json_doc = self._map_file_obj_to_json_document(file)
                docs_not_in_db.append((str(file.id), json_doc))

        return file_data, docs_not_in_db

    def _add_data_from_doc_to_file(self, file: File, doc: JsonDocument) -> File:
        json_dict = json.loads(doc.json_value)
        for metadata_field in self.METADATA_FIELD_SET:
            value = json_dict.get(metadata_field)
            if not value:
                logger.debug(
                    f"The file metadata field {metadata_field} is not in the respective json document."
                )
                continue
            file.__setattr__(metadata_field, value)
        return file

    def _delete_file_versions(
        self, project_id: str, file_key: str, keep_latest_version: bool
    ) -> None:
        bucket_name = get_bucket_name(project_id, self.sys_namespace)
        files_to_prefix = self.list_files(
            project_id, recursive=False, include_versions=True, prefix=file_key
        )

        delete_objects = []
        latest_version_db_key = ""
        for file in files_to_prefix:
            if file.key != file_key:
                # This might happen, since list_files takes only a prefix
                continue
            if keep_latest_version and file.latest_version:
                latest_version_db_key = generate_file_id(file_key, str(file.version))
                continue

            delete_objects.append(DeleteObject(file_key, file.version))

        if not delete_objects:
            logger.debug(f"No versions for deletion (file_key: {file_key}).")
            return

        # TODO: Validate bypass_governance_mode behavior
        errors = self.client.remove_objects(
            bucket_name, delete_objects, bypass_governance_mode=True
        )

        for error in errors:
            # TODO: Critical is only for testing an should be handled when tested
            logger.critical(
                f"Error removing old versions of file {file_key} (Errors: {error}).",
            )

        # Check related versions against the DB so that everything gets cleaned up
        json_path_filter = f'$ ? (@.key == "{file_key}")'
        db_keys = [
            doc.key
            for doc in self._json_db_manager.list_json_documents(
                project_id, self.DOC_COLLECTION_NAME, json_path_filter
            )
        ]
        if keep_latest_version:
            db_keys.remove(latest_version_db_key)

        for db_key in db_keys:
            self._json_db_manager.delete_json_document(
                project_id, self.DOC_COLLECTION_NAME, db_key
            )



import string
import time
from typing import Any, Dict, List, Optional, Tuple, Union

from kubernetes import client as kube_client
from kubernetes.client.models import (
    V1Container,
    V1Deployment,
    V1DeploymentSpec,
    V1EnvVar,
    V1Job,
    V1LabelSelector,
    V1NetworkPolicy,
    V1NetworkPolicyIngressRule,
    V1NetworkPolicyPeer,
    V1NetworkPolicySpec,
    V1ObjectMeta,
    V1PersistentVolumeClaim,
    V1PersistentVolumeClaimSpec,
    V1PersistentVolumeClaimVolumeSource,
    V1Pod,
    V1PodList,
    V1PodSpec,
    V1PodTemplateSpec,
    V1ResourceRequirements,
    V1Service,
    V1ServicePort,
    V1ServiceSpec,
    V1Volume,
    V1VolumeMount,
)
from kubernetes.client.rest import ApiException

from contaxy.config import settings
from contaxy.managers.deployment.utils import (
    _MIN_MEMORY_DEFAULT_MB,
    Labels,
    get_label_string,
    get_project_selection_labels,
    map_labels,
    map_list_to_string,
)
from contaxy.schema import Job, Service
from contaxy.schema.deployment import (
    Deployment,
    DeploymentCompute,
    DeploymentStatus,
    DeploymentType,
)
from contaxy.schema.exceptions import ResourceNotFoundError, ServerBaseError
from contaxy.utils.utils import remove_none_values_from_dict


def get_label_selector(label_pairs: List[Tuple[str, str]]) -> str:
    """Bring label tuples into the form required by the Kubernetes client, e.g. 'key1=value1,key2=value2,key3=value3)'."""
    return ",".join([get_label_string(pair[0], pair[1]) for pair in label_pairs])


def get_deployment_selection_labels(
    project_id: str, deployment_type: DeploymentType = DeploymentType.SERVICE
) -> str:
    """Return selector identifying project services/jobs in the form Kubernetes expects a label selector.

    Args:
        project_id (str): The project id of the resources to select.
        deployment_type (DeploymentType, optional): The deployment type by which the selected resources are filtered. Defaults to DeploymentType.SERVICE.

    Returns:
        str: Kubernetes label string in the form of 'key1=value1,key2=value2,key3=value3,...'
    """
    return get_label_selector(
        get_project_selection_labels(
            project_id=project_id, deployment_type=deployment_type
        )
    )


# TODO: Return list of pods? As there can be multiple pods belonging to the same job / deployment (e.g. which were created / failed but do not run anymore)
def get_pod(
    project_id: str,
    service_id: str,
    kube_namespace: str,
    core_api: kube_client.CoreV1Api,
) -> Optional[V1Pod]:
    """Get the pod filtered by the project id and service id labels in the given Kubernetes namespace.

    Args:
        service_id (str): If deployed via Contaxy, corresponds to the deployment id of the pod.
        kube_namespace (str): The Kubernetes namespaces in which to look for the pod.
        core_api (kube_client.CoreV1Api): Initialized Kubernetes CoreV1Api object.

    Raises:
        ResourceNotFoundError: Raised when no pod matches the selection criteria.

    Returns:
        Optional[V1Pod]: Returns the pod matching the selection criteria. In case of replicas, multiple pods can match the criteria; in this case, the first pod is selected arbitrarily.
    """
    label_selector = get_label_selector(
        [
            (Labels.NAMESPACE.value, settings.SYSTEM_NAMESPACE),
            (Labels.PROJECT_NAME.value, project_id),
            (Labels.DEPLOYMENT_ID.value, service_id),
        ]
    )

    pods: V1PodList = core_api.list_namespaced_pod(
        namespace=kube_namespace, label_selector=label_selector
    )

    if len(pods.items) == 0:
        raise ResourceNotFoundError(
            f"No pod with label-selector {label_selector} found"
        )

    return pods.items[0]


def create_pvc(
    pvc: Optional[V1PersistentVolumeClaim],
    kube_namespace: str,
    core_api: kube_client.CoreV1Api,
) -> None:
    if pvc is None:
        return

    try:
        core_api.create_namespaced_persistent_volume_claim(
            namespace=kube_namespace, body=pvc
        )
    except ApiException as e:
        # 409 means that the pvc was not created but already exists and can be used
        if e.status != 409:
            raise ServerBaseError(
                f"Could not create persistent volume claim for service '{pvc.metadata.name}' with reason: {e}"
            )


def create_service(
    service_config: Optional[V1Service],
    kube_namespace: str,
    core_api: kube_client.CoreV1Api,
) -> None:
    if service_config is None:
        return

    try:
        core_api.create_namespaced_service(
            namespace=kube_namespace, body=service_config
        )
    except ApiException as e:
        raise ServerBaseError(
            f"Could not create namespaced service '{service_config.metadata.name}' with reason: {e}"
        )


def build_kube_service_config(
    service: Service, project_id: str, kube_namespace: str
) -> V1Service:
    # TODO: set the endpoints as annotations
    service_ports: Dict[str, V1ServicePort] = {}
    if service.endpoints:
        for endpoint in service.endpoints:

            # An endpoint can be in the form of one of the examples ["8080", "9001/webapp/ui", "9002b"]. See the Contaxy endpoint docs <LINK> for more information.
            port_number = endpoint.split("/")[0].replace("b", "")

            # Kubernetes throws an error when the same port is defined for the same service
            if port_number in service_ports:
                continue

            service_port = V1ServicePort(
                name=port_number, protocol="TCP", port=int(port_number)
            )
            service_ports[port_number] = service_port

    return V1Service(
        metadata=V1ObjectMeta(
            namespace=kube_namespace,
            name=service.id,
            labels={
                # Labels.DISPLAY_NAME.value: service.display_name,# display_name cannot be a label since whitespaces are not allowed in label values. Kubernetes validation regex: (([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?')
                Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                Labels.PROJECT_NAME.value: project_id,
                Labels.DEPLOYMENT_ID.value: service.id,
                Labels.DEPLOYMENT_TYPE.value: service.deployment_type,
            },  # service.labels
        ),
        spec=V1ServiceSpec(
            selector={
                Labels.DEPLOYMENT_ID.value: service.id,
                Labels.PROJECT_NAME.value: project_id,
                Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                Labels.DEPLOYMENT_TYPE.value: service.deployment_type,
            },
            # ports must be set and it must contain at least one port
            ports=list(service_ports.values())
            if len(service_ports.values()) > 0
            else [V1ServicePort(name="default", protocol="TCP", port=80)],
            type="ClusterIP",
        ),
    )


def build_pod_template_spec(
    deployment: Deployment,
    metadata: V1ObjectMeta,
) -> V1PodTemplateSpec:
    compute_resources = deployment.compute
    # TODO: check default values and store them globally probably!
    min_cpus = compute_resources.min_cpus or 0
    max_cpus = compute_resources.max_cpus or 1
    min_memory = compute_resources.min_memory or _MIN_MEMORY_DEFAULT_MB
    max_memory = compute_resources.max_memory or 100
    max_container_size = compute_resources.max_container_size or 100
    resource_requirements = V1ResourceRequirements(
        limits={
            # e.g. 8000m = 8000 micro-cores = 8 cores (https://cloud.google.com/blog/products/gcp/kubernetes-best-practices-resource-requests-and-limits)
            "cpu": f"{max_cpus * 1000}m",
            "memory": f"{max_memory}M",
            "ephemeral-storage": f"{max_container_size}M",
        },
        requests={
            "cpu": f"{min_cpus * 1000}m",
            "memory": f"{min_memory}M",
            "ephemeral-storage": "0M",
        },
    )

    # the name is used by Kubernetes to match the container-volume and the pod-volume section
    mount_name = f"mount-{deployment.id}"
    volume_mounts, volumes = None, None
    if compute_resources.volume_path:
        volume_mounts = [
            V1VolumeMount(
                name=mount_name, mount_path=str(compute_resources.volume_path)
            )
        ]
        volumes = [
            V1Volume(
                name=mount_name,
                persistent_volume_claim=V1PersistentVolumeClaimVolumeSource(
                    claim_name=deployment.id
                ),
            )
        ]
    container = V1Container(
        name=deployment.id,
        image=deployment.container_image,
        image_pull_policy="IfNotPresent",
        resources=resource_requirements,
        command=deployment.command,
        args=deployment.args,
        env=[
            V1EnvVar(name=name, value=value)
            for name, value in deployment.parameters.items()
        ],
        volume_mounts=volume_mounts,
    )
    pod_spec = V1PodSpec(
        volumes=volumes,
        containers=[container],
    )

    return V1PodTemplateSpec(metadata=metadata, spec=pod_spec)


def build_deployment_metadata(
    kube_namespace: str,
    project_id: str,
    deployment: Deployment,
) -> V1ObjectMeta:
    return V1ObjectMeta(
        namespace=kube_namespace,
        name=deployment.id,
        labels={
            Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
            Labels.PROJECT_NAME.value: project_id,
            Labels.DEPLOYMENT_ID.value: deployment.id,
            Labels.DEPLOYMENT_TYPE.value: deployment.deployment_type.value,
        },
        annotations={
            **deployment.metadata,
            Labels.DISPLAY_NAME.value: deployment.display_name,
            Labels.DEPLOYMENT_TYPE.value: deployment.deployment_type.value,
            Labels.DESCRIPTION.value: deployment.description,
            Labels.ENDPOINTS.value: map_list_to_string(deployment.endpoints),
            Labels.REQUIREMENTS.value: map_list_to_string(deployment.requirements),
            Labels.ICON.value: deployment.icon,
            Labels.MIN_LIFETIME.value: str(deployment.compute.min_lifetime),
            Labels.CREATED_BY.value: deployment.created_by,
            Labels.VOLUME_PATH.value: deployment.compute.volume_path,
        },
    )


def build_kube_deployment_config(
    service: Service,
    project_id: str,
    kube_namespace: str,
) -> Tuple[V1Deployment, Union[V1PersistentVolumeClaim, None]]:
    compute_resources = service.compute

    metadata = build_deployment_metadata(
        kube_namespace=kube_namespace,
        project_id=project_id,
        deployment=service,
    )

    pvc = None
    # add mount - handle PVC; ignore Secret, Bind and NFS for now
    if compute_resources.volume_path:
        # TODO: change the default storage size
        default_storage_size = "100M"
        storage_size = (
            f"{compute_resources.max_volume_size}M"
            if compute_resources.max_volume_size is not None
            and compute_resources.max_volume_size > 0
            else default_storage_size
        )
        pvc = V1PersistentVolumeClaim(
            metadata=metadata,
            spec=V1PersistentVolumeClaimSpec(
                # TODO: make the storage class name configurable?
                storage_class_name="default",
                resources=V1ResourceRequirements(requests={"storage": storage_size}),
                access_modes=["ReadWriteOnce"],
            ),
        )

    # TODO: adding of ServiceAccountName needed?

    # --

    deployment = V1Deployment(
        metadata=metadata,
        spec=V1DeploymentSpec(
            # TODO: the max_replicas name does not make so much sense. Probably just replicas?
            replicas=compute_resources.max_replicas
            if compute_resources.max_replicas
            else 1,
            selector=V1LabelSelector(
                match_labels={
                    Labels.DEPLOYMENT_ID.value: service.id,
                    Labels.PROJECT_NAME.value: project_id,
                    Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                    Labels.DEPLOYMENT_TYPE.value: service.deployment_type.value,
                }
            ),
            template=build_pod_template_spec(
                deployment=service,
                metadata=metadata,
            ),
            # TODO: Make min ready seconds configurable? (see https://kubernetes.io/docs/concepts/workloads/controllers/deployment/#min-ready-seconds)
            min_ready_seconds=5,
        ),
    )

    return deployment, pvc


def build_project_network_policy_spec(
    project_id: str, kube_namespace: str
) -> V1NetworkPolicy:
    project_pod_selector = V1LabelSelector(
        match_labels={
            Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
            Labels.PROJECT_NAME.value: project_id,
        }
    )

    network_policy = V1NetworkPolicy(
        metadata=V1ObjectMeta(
            namespace=kube_namespace,
            name=project_id,
            labels={
                Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                Labels.PROJECT_NAME.value: project_id,
            },
        ),
        spec=V1NetworkPolicySpec(
            ingress=[
                V1NetworkPolicyIngressRule(
                    # Allow traffic from other pods of the same project and the Contaxy backend container
                    _from=[
                        V1NetworkPolicyPeer(pod_selector=project_pod_selector),
                        # TODO: probably not needed as this rule is added via our core kube-network-policy
                        V1NetworkPolicyPeer(
                            pod_selector=V1LabelSelector(
                                match_labels={
                                    Labels.NAMESPACE.value: settings.SYSTEM_NAMESPACE,
                                    Labels.DEPLOYMENT_TYPE.value: f"{DeploymentType.CORE_BACKEND.value}",
                                }
                            )
                        ),
                    ]
                )
            ],
            pod_selector=project_pod_selector,
        ),
    )

    return network_policy


def check_or_create_project_network_policy(
    network_policy: V1NetworkPolicy, networking_api: kube_client.NetworkingV1Api
) -> V1NetworkPolicy:
    try:
        return networking_api.read_namespaced_network_policy(
            namespace=network_policy.metadata.namespace,
            name=network_policy.metadata.name,
        )
    except ApiException:
        return networking_api.create_namespaced_network_policy(
            namespace=network_policy.metadata.namespace, body=network_policy
        )


def wait_for_deployment(
    deployment_name: str,
    kube_namespace: str,
    apps_api: kube_client.AppsV1Api,
    timeout: int = 60,
) -> None:
    start = time.time()
    while time.time() - start < timeout:
        time.sleep(2)
        response = apps_api.read_namespaced_deployment_status(
            namespace=kube_namespace, name=deployment_name
        )
        s = response.status
        if (
            s.updated_replicas == response.spec.replicas
            and s.replicas == response.spec.replicas
            and s.available_replicas == response.spec.replicas
            and s.observed_generation >= response.metadata.generation
        ):
            return
        else:
            # print(
            #     f"[updated_replicas:{s.updated_replicas},replicas:{s.replicas}"
            #     ",available_replicas:{s.available_replicas},observed_generation:{s.observed_generation}] waiting..."
            # )
            pass

    raise RuntimeError(f"Waiting timeout for deployment {deployment_name}")


def wait_for_job(
    job_name: str,
    kube_namespace: str,
    batch_api: kube_client.BatchV1Api,
    timeout: int = 60,
) -> None:
    start = time.time()
    while time.time() - start < timeout:
        response = batch_api.read_namespaced_job_status(
            namespace=kube_namespace, name=job_name
        )
        s = response.status
        if (s.active and s.active > 0) or (s.succeeded and s.succeeded > 0):
            return

    raise RuntimeError(f"Waiting timeout for job {job_name}")


def wait_for_deletion(
    api: Union[kube_client.AppsV1Api, kube_client.BatchV1Api],
    kube_namespace: str,
    deployment_id: str,
) -> None:
    start = time.time()
    timeout = 60
    while time.time() - start < timeout:
        try:
            if type(api) == kube_client.AppsV1Api:
                api.read_namespaced_deployment_status(
                    namespace=kube_namespace, name=deployment_id
                )
            elif type(api) == kube_client.BatchV1Api:
                api.read_namespaced_job(name=deployment_id, namespace=kube_namespace)
            time.sleep(2)
        except ApiException:
            # Deployment is deleted
            break


def map_deployment(deployment: Union[V1Deployment, V1Job]) -> Dict[str, Any]:
    # We assume a 1:1 mapping between deployment/pod & containers
    container = deployment.spec.template.spec.containers[0]
    mapped_labels = map_labels(
        {**deployment.metadata.labels, **deployment.metadata.annotations}
    )

    resources = container.resources
    compute_resources = DeploymentCompute(
        min_cpus=int(resources.requests["cpu"].strip(string.ascii_letters)),  # / 1e9,
        max_cpus=int(resources.limits["cpu"].strip(string.ascii_letters)),  # / 1e9,
        min_memory=int(
            resources.requests["memory"].strip(string.ascii_letters)
        ),  # / 1000 / 1000,
        max_memory=int(
            resources.limits["memory"].strip(string.ascii_letters)
        ),  # / 1000 / 1000,
        max_gpus=None,  # TODO: fill with sensible information - where to get it from?
        min_lifetime=mapped_labels.min_lifetime or 0,
        volume_path=mapped_labels.volume_path,
        # TODO: add max_volume_size, max_replicas
    )

    try:
        if deployment.metadata.deletion_timestamp is not None:
            status = DeploymentStatus.TERMINATING.value
        else:
            # TODO: detect failed?
            # TODO: return RUNNING when at least MIN_REPLICAS is fulfilled
            # TODO: if no replica is started, return FAILED
            # TODO: if READY_REPLICAS < MIN_REPLICAS => check exit code of POD
            # TODO: set failed when creation_timestamp is high but still unavailable
            successful = (
                deployment.status.ready_replicas == deployment.status.replicas
                if hasattr(deployment.status, "ready_replicas")
                else deployment.status.succeeded
            )
            if successful:
                status = DeploymentStatus.RUNNING.value
            else:
                status = DeploymentStatus.PENDING.value
    except ValueError:
        status = DeploymentStatus.UNKNOWN.value

    started_at = deployment.metadata.creation_timestamp
    # stopped_at only relevant for Jobs
    stopped_at = (
        deployment.status.completion_time
        if hasattr(deployment.status, "completion_time")
        else 0
    )

    parameters = {}
    for env in container.env:
        parameters[env.name] = env.value if env.value else ""

    return {
        "container_image": container.image,
        "command": container.command,
        "args": container.args,
        "metadata": mapped_labels.metadata,
        "compute": compute_resources,
        "deployment_type": mapped_labels.deployment_type,
        "description": mapped_labels.description,
        # TODO: replicase __ logic => use annotations instead of labels for all metadata that we don't use for filtering
        "display_name": mapped_labels.display_name.replace("__", " ")
        if mapped_labels.display_name
        else "",
        "endpoints": mapped_labels.endpoints,
        # TODO: exit_code can only be queried on pod-level, but what if there are multiple replicas?
        # "exit_code": container.attrs.get("State", {}).get("ExitCode", -1),
        "icon": mapped_labels.icon,
        "id": deployment.metadata.name,
        "internal_id": f"{deployment.metadata.name}-{deployment.metadata.uid}",
        "parameters": parameters,
        "started_at": started_at,
        "status": status,
        "stopped_at": stopped_at,
    }


def map_kube_service(
    deployment: V1Deployment,
) -> Service:
    transformed_deployment = map_deployment(deployment=deployment)
    transformed_deployment = remove_none_values_from_dict(transformed_deployment)
    return Service(**transformed_deployment)


def map_kube_job(job: V1Job) -> Job:
    transformed_job = map_deployment(deployment=job)
    transformed_job = remove_none_values_from_dict(transformed_job)
    # TODO: add status SUCCESSFUL or FAILED
    return Job(**transformed_job)



from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, List, Optional

from fastapi import Path
from pydantic import BaseModel, Field

from contaxy.schema.shared import Resource, ResourceInput

SERVICE_ID_PARAM = Path(
    ...,
    title="Service ID",
    description="A valid Service ID.",
    # TODO: add length restriction
)

JOB_ID_PARAM = Path(
    ...,
    title="Job ID",
    description="A valid job ID.",
    # TODO: add length restriction
)


class DeploymentType(str, Enum):
    CORE_BACKEND = "core-backend"
    SERVICE = "service"
    JOB = "job"
    EXTENSION = "extension"
    UNKNOWN = "unknown"


class DeploymentStatus(str, Enum):
    # Deployment created, but not ready for usage
    PENDING = "pending"  # Alternative naming: waiting
    # Deployment is running and usable
    RUNNING = "running"
    # Deployment was stopped with successful exit code (== 0).
    SUCCEEDED = "succeeded"
    # Deployment was stopped with failure exit code (> 0).
    FAILED = "failed"
    # Deployment was deleted and is now terminating the pods.
    TERMINATING = "terminating"
    # Deployment is stopped (it still exists in the DB but no container is running)
    STOPPED = "stopped"
    # Deployment state cannot be obtained.
    UNKNOWN = "unknown"
    # Deployment is paused (only on docker?)/
    # PAUSED = "paused"
    # Other possible options:
    # KILLED = "killed"
    # STARTING("starting"),
    # STOPPPED("stopped"),
    # CREATED("created"),
    # REBOOTING
    # TERMINATED: Container is terminated. This container can’t be started later on.
    # STOPPED:  Container is stopped. This container can be started later on.
    # ERROR – Container is an error state. Usually no operations can be performed on the container once it ends up in the error state.
    # SUSPENDED: Container is suspended.


class DeploymentCompute(BaseModel):
    min_cpus: Optional[int] = Field(
        None,
        example=2,
        ge=0,
        description="Minimum number of CPU cores required by this deployment. The system will make sure that atleast this amount is available to the deployment.",
    )
    max_cpus: Optional[int] = Field(
        None,
        example=4,
        ge=0,
        description="Maximum number of CPU cores. Even so the system will try to provide the specified amount, it's only guaranteed that the deployment cannot use more. 0 means unlimited.",
    )
    min_memory: Optional[int] = Field(
        None,
        example=4000,
        ge=5,  # 4 is the minimal RAM needed for containers
        description="Minimum amount of memory in Megabyte required by this deployment. The system will make sure that atleast this amount is available to the deployment.",
    )
    max_memory: Optional[int] = Field(
        None,
        example=8000,
        ge=0,
        description="Maximum amount of memory in Megabyte. Even so the system will try to provide the specified amount, it's only guaranteed that the deployment cannot use more. 0 means unlimited.",
    )  # in MB
    min_gpus: Optional[int] = Field(
        None,
        example=1,
        ge=0,
        description="Minimum number of GPUs required by this deployments. The system will make sure that atleast this amount is available to the deployment.",
    )
    max_gpus: Optional[int] = Field(
        None,
        example=2,
        ge=0,
        description="Maximum number of GPUs. Even so the system will try to provide the specified amount, it's only guaranteed that the deployment cannot use more.",
    )
    volume_path: Optional[str] = Field(
        None,
        example="/path/to/data",
        description="Container internal directory that should mount a volume for data persistence.",
    )
    # TODO: min_volume_size
    max_volume_size: Optional[int] = Field(
        None,
        example=32000,
        ge=1,
        description="Maximum volume size in Megabyte. This is only applied in combination with volume_path.",
    )
    # TODO: min_container_size
    max_container_size: Optional[int] = Field(
        None,
        example=32000,
        ge=1,
        description="Maximum container size in Megabyte. The deployment will be killed if it grows above this limit.",
    )
    # TODO: min_replicas
    max_replicas: int = Field(
        1,
        example=2,
        ge=1,
        description="Maximum number of deployment instances. The system will make sure to optimize the deployment based on the available resources and requests. Use 1 if the deployment is not scalable.",
    )
    # TODO: use timedelta
    min_lifetime: int = Field(
        0,
        example=86400,
        description="Minimum guaranteed lifetime in seconds. Once the lifetime is reached, the system is allowed to kill the deployment in case it requires additional resources.",
    )


class DeploymentBase(BaseModel):
    container_image: str = Field(
        ...,
        example="hello-world:latest",
        max_length=2000,
        description="The container image used for this deployment.",
    )
    parameters: Dict[str, str] = Field(
        {},
        example={"TEST_PARAM": "param-value"},
        description="Parameters (environment variables) for this deployment.",
    )
    compute: DeploymentCompute = Field(
        DeploymentCompute(),
        description="Compute instructions and limitations for this deployment.",
    )
    command: Optional[List[str]] = Field(
        None,
        description="Command to run within the deployment. This overwrites the existing docker ENTRYPOINT.",
    )
    args: Optional[List[str]] = Field(
        None,
        description="Arguments to the command/entrypoint. This overwrites the existing docker CMD.",
    )
    requirements: List[str] = Field(
        [],
        description="Additional requirements for deployment.",
    )
    endpoints: List[str] = Field(
        [],
        example=["8080", "9001/webapp/ui", "9002b"],
        description="A list of HTTP endpoints that can be accessed. This should always have an internal port and can include additional instructions, such as the URL path.",
    )
    # TODO: v2
    # input_files: Optional[List[dict]] = Field(
    #    None,
    #    description="A list of files that should be added to the deployment.",
    # )
    # TODO: v2
    # command_args: Optional[List[str]] = Field(
    #    None,
    #    description="Arguments to use for the command of the deployment. This overwrites the existing arguments.",
    # )


class DeploymentInput(ResourceInput, DeploymentBase):
    pass


class Deployment(Resource, DeploymentBase):
    started_at: Optional[datetime] = Field(
        None,
        description="Timestamp when the deployment was started.",
    )
    stopped_at: Optional[datetime] = Field(
        None,
        description="Timestamp when the container has stopped.",
    )
    extension_id: Optional[str] = Field(
        None,
        description="The extension ID in case the deployment is deployed via an extension.",
    )
    deployment_type: DeploymentType = Field(
        DeploymentType.UNKNOWN,
        description="The type of this deployment.",
    )
    status: DeploymentStatus = Field(
        DeploymentStatus.UNKNOWN,
        example=DeploymentStatus.RUNNING,
        description="The status of this deployment.",
    )
    internal_id: Optional[str] = Field(
        None,
        example="73d247087fea5bfb3a67e98da6a07f5bf4e2a90e5b52f3c12875a35600818376",
        description="The ID of the deployment on the orchestration platform.",
    )
    # TODO: All labels should be transformed into the metadata or additional_metadata
    # deployment_labels: Optional[Dict[str, str]] = Field(
    #    None,
    #    example={"foo.bar.label": "label-value"},
    #    description="The labels of the deployment on the orchestration platform.",
    # )
    # TODO: should be a debug information.
    # exit_code: Optional[int] = Field(
    #    None,
    #    example=0,
    #    description="The Exit code of the container, in case the container was stopped.",
    # )


class ServiceBase(BaseModel):
    graphql_endpoint: Optional[str] = Field(
        None,
        example="8080/graphql",
        description="GraphQL endpoint.",
    )
    openapi_endpoint: Optional[str] = Field(
        None,
        example="8080/openapi.yaml",
        description="Endpoint that prorvides an OpenAPI schema definition..",
    )
    health_endpoint: Optional[str] = Field(
        None,
        example="8080/healthz",
        description="The endpoint instruction that can be used for checking the deployment health.",
    )
    idle_timeout: Optional[timedelta] = Field(
        None,
        description="Time after which the service is considered idling and will be stopped during the next idle check."
        "If set to None, the service will never be considered idling."
        "Can be specified as seconds or ISO 8601 time delta.",
    )
    clear_volume_on_stop: bool = Field(
        False,
        description="If set to true, any volume attached to the service be deleted when the service is stopped."
        "This means all persisted data will be cleared on service stop.",
    )


class ServiceInput(ServiceBase, DeploymentInput):
    is_stopped: bool = Field(
        False,
        description="If set to true, the service will be created in the DB but not started. The service status will be 'stopped'.",
    )


class ServiceUpdate(ServiceInput):
    # Add default value for container_image so it is not required to be set in an update request
    container_image: str = Field(
        "",
        example="hello-world:latest",
        max_length=2000,
        description="The container image used for this deployment.",
    )
    # Allow None for parameters and metadata values so they can be completely removed in an update request
    parameters: Dict[str, Optional[str]] = Field(  # type: ignore[assignment]
        {},
        example={"TEST_PARAM": "param-value"},
        description="Parmeters (enviornment variables) for this deployment.",
    )
    metadata: Dict[str, Optional[str]] = Field(  # type: ignore[assignment]
        {},
        example={"additional-metadata": "value"},
        description="A collection of arbitrary key-value pairs associated with this resource that does not need predefined structure. Enable third-party integrations to decorate objects with additional metadata for their own use.",
    )


class Service(ServiceBase, Deployment):
    # Store the last time this service was accessed and by which user
    last_access_time: Optional[datetime] = Field(
        None, description="Timestamp of the last time the service was accessed."
    )
    last_access_user: Optional[str] = Field(
        None, description="Id of the user that last accessed the service."
    )


class JobBase(BaseModel):
    pass


class JobInput(JobBase, DeploymentInput):
    pass
    # TODO: v2
    # output_files: Optional[List[dict]] = Field(
    #    None,
    #    description="A list of container internal files that should be uploaded to the storage once the job has succeeded.",
    # )


class Job(JobBase, Deployment):
    pass



"""Data Models and Schemas."""


from .auth import (
    AccessLevel,
    ApiToken,
    AuthorizedAccess,
    OAuth2TokenRequestFormNew,
    OAuthToken,
    OAuthTokenIntrospection,
    TokenType,
    User,
    UserInput,
    UserRegistration,
)
from .deployment import (
    Deployment,
    DeploymentInput,
    Job,
    JobInput,
    Service,
    ServiceInput,
)
from .exceptions import (
    AUTH_ERROR_RESPONSES,
    ClientBaseError,
    ClientValueError,
    PermissionDeniedError,
    ResourceAlreadyExistsError,
    ResourceNotFoundError,
    ResourceUpdateFailedError,
    ServerBaseError,
    UnauthenticatedError,
)
from .extension import Extension, ExtensionInput
from .file import File, FileInput, FileStream
from .json_db import JsonDocument
from .project import Project, ProjectCreation, ProjectInput
from .shared import CoreOperations, ExtensibleOperations, ResourceAction
from .system import SystemInfo, SystemStatistics



"""Utilities for managing global and request state for an FastAPI app."""

from typing import Any, Callable, Optional

import addict
from starlette import datastructures

from contaxy.config import Settings
from contaxy.schema import AuthorizedAccess


class _Dict(addict.Dict):
    def __missing__(self, key):  # type: ignore
        # Do not create empty dict objects (default)
        return None


class State:
    _SHARED_NAMESPACE_KEY = "shared"

    def __init__(self, state: datastructures.State):
        """Initializes the state."""
        # Add namespaces attribute to state
        if not hasattr(state, "namespaces"):
            state.namespaces = {}
        self._namespaces = state.namespaces

        # Add close callbacks attribute to state
        if not hasattr(state, "_close_callbacks"):
            state._close_callbacks = []
        self._close_callbacks = state._close_callbacks

    @property
    def namespaces(self) -> dict:
        return self._namespaces

    def __getitem__(self, namespace: Any) -> addict.Dict:
        if namespace not in self._namespaces:
            self._namespaces[namespace] = _Dict()
        return self._namespaces[namespace]

    def __setitem__(self, namespace: Any, value: Any) -> None:
        if namespace == State._SHARED_NAMESPACE_KEY:
            raise ValueError("The shared namespace is not allowed to be set.")
        self._namespaces[namespace] = value

    def __delitem__(self, namespace: Any) -> None:
        if namespace in self._namespaces:
            del self._namespaces[namespace]

    def __enter__(self) -> "State":
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore
        self.close()

    @property
    def shared_namespace(self) -> addict.Dict:
        return self[State._SHARED_NAMESPACE_KEY]

    def register_close_callback(self, callback_func: Callable) -> None:
        """Registers a close callback.

        All registered callback functions are called when the state is closed.
        """
        self._close_callbacks.append(callback_func)

    def close(self) -> None:
        """Closes state.

        This will execute all registered close callback functions.
        This method is automatically called. Manual calls should only be done under careful consideration.
        """
        if self._close_callbacks:
            for close_callback in self._close_callbacks:
                close_callback()
        self._close_callbacks.clear()


class GlobalState(State):
    """Holds a global state of one app instance (process).

    The global state is created once for every app instance/process
    and can be used to store and share objects globally (between all components),
    such as DB connections, HTTP clients, or data caches.
    """

    _SETTINGS_KEY = "settings"

    def __init__(self, state: datastructures.State):
        """Initializes the global state.

        Args:
            state: The state object from the app (`app.state`)
        """
        super(GlobalState, self).__init__(state)

    @property
    def settings(self) -> Settings:
        """Returns the global platform settings."""
        return self.shared_namespace.settings

    @settings.setter
    def settings(self, settings: Settings) -> None:
        """Sets the global platform settings.

        This is automatically set at the start of the app instance/process.
        """
        self.shared_namespace.settings = settings


class RequestState(State):
    """Holds a state for a single request.

    The request state is created once for every request
    and can be used to store and share objects between all components,
    such as DB connections, HTTP clients, or data caches.
    """

    def __init__(self, state: datastructures.State):
        """Initializes the request state.

        Args:
            state: The state object from the request (`request.state`)
        """
        super(RequestState, self).__init__(state)

    @property
    def authorized_access(self) -> Optional[AuthorizedAccess]:
        """Returns the authorized access info for the request."""
        return self.shared_namespace.authorized_access

    @authorized_access.setter
    def authorized_access(self, authorized_access: AuthorizedAccess) -> None:
        """Sets the authorized access info."""
        self.shared_namespace.authorized_access = authorized_access

    @property
    def authorized_subject(self) -> str:
        """Convenience method to check for authorized_access being not None and returns the authorized_subject."""

        if self.authorized_access and self.authorized_access.authorized_subject:
            return self.authorized_access.authorized_subject
        else:
            return ""



from datetime import datetime
from typing import Dict, List, Literal, Optional

import requests
from pydantic.tools import parse_raw_as

from contaxy.clients.shared import handle_errors
from contaxy.operations.deployment import DeploymentOperations
from contaxy.schema import Job, JobInput, ResourceAction, Service, ServiceInput
from contaxy.schema.deployment import DeploymentType, ServiceUpdate
from contaxy.schema.shared import ResourceActionExecution


class DeploymentManagerClient(DeploymentOperations):
    def __init__(self, client: requests.Session):
        self._client = client

    @property
    def client(self) -> requests.Session:
        return self._client

    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
        request_kwargs: Dict = {},
    ) -> List[Service]:
        response = self.client.get(f"/projects/{project_id}/services", **request_kwargs)
        if deployment_type == DeploymentType.EXTENSION:
            raise ValueError("Use the ExtensionManager to list extensions!")
        handle_errors(response)
        return parse_raw_as(List[Service], response.text)

    def deploy_service(
        self,
        project_id: str,
        service_input: ServiceInput,
        action_id: Optional[str] = None,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
        wait: bool = False,
        request_kwargs: Dict = {},
    ) -> Service:
        params = {}
        if wait:
            params["wait"] = "true"
        if action_id:
            params["action_id"] = action_id
        if deployment_type == DeploymentType.EXTENSION:
            raise ValueError("Use the ExtensionManager to deploy extensions!")

        resource = self.client.post(
            f"/projects/{project_id}/services",
            params=params,
            data=service_input.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(Service, resource.text)

    def update_service(
        self,
        project_id: str,
        service_id: str,
        service: ServiceUpdate,
        request_kwargs: Dict = {},
    ) -> Service:
        resource = self.client.patch(
            f"/projects/{project_id}/services/{service_id}",
            data=service.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(Service, resource.text)

    def update_service_access(
        self,
        project_id: str,
        service_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        resource = self.client.post(
            f"/projects/{project_id}/services/{service_id}:update-service-access",
            **request_kwargs,
        )
        handle_errors(resource)

    def list_deploy_service_actions(
        self,
        project_id: str,
        service: ServiceInput,
        request_kwargs: Dict = {},
    ) -> List[ResourceAction]:
        resources = self.client.post(
            f"/projects/{project_id}/services:deploy-actions",
            data=service.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resources)
        return parse_raw_as(List[ResourceAction], resources.text)

    def get_service_metadata(
        self,
        project_id: str,
        service_id: str,
        request_kwargs: Dict = {},
    ) -> Service:
        resource = self.client.get(
            f"/projects/{project_id}/services/{service_id}", **request_kwargs
        )
        handle_errors(resource)
        return parse_raw_as(Service, resource.text)

    def delete_service(
        self,
        project_id: str,
        service_id: str,
        delete_volumes: bool = False,
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.delete(
            f"/projects/{project_id}/services/{service_id}",
            params={"delete_volumes": delete_volumes},
            **request_kwargs,
        )
        handle_errors(response)

    def delete_services(
        self,
        project_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.delete(
            f"/projects/{project_id}/services",
            **request_kwargs,
        )
        handle_errors(response)

    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int],
        since: Optional[datetime],
        request_kwargs: Dict = {},
    ) -> str:
        params = {}
        if lines:
            params["lines"] = str(lines)
        if since:
            params["since"] = since.__str__()
        response = self.client.get(
            f"/projects/{project_id}/services/{service_id}/logs",
            params=params,
            **request_kwargs,
        )
        handle_errors(response)
        return response.json()

    def suggest_service_config(
        self,
        project_id: str,
        container_image: str,
        request_kwargs: Dict = {},
    ) -> ServiceInput:
        resource = self.client.get(
            f"/projects/{project_id}/services:suggest-config",
            params={"container_image": container_image},
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(ServiceInput, resource.text)

    def list_service_actions(
        self,
        project_id: str,
        service_id: str,
        request_kwargs: Dict = {},
    ) -> List[ResourceAction]:
        resources = self.client.get(
            f"/projects/{project_id}/services/{service_id}/actions", **request_kwargs
        )
        handle_errors(resources)
        return parse_raw_as(List[ResourceAction], resources.text)

    def execute_service_action(
        self,
        project_id: str,
        service_id: str,
        action_id: str,
        action_execution: ResourceActionExecution = ResourceActionExecution(),
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.post(
            f"/projects/{project_id}/services/{service_id}/actions/{action_id}",
            data=action_execution.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(response)
        # TODO: Return response?

    def access_service(
        self,
        project_id: str,
        service_id: str,
        endpoint: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.get(
            f"/projects/{project_id}/services/{service_id}/access/{endpoint}",
            **request_kwargs,
        )
        handle_errors(response)
        # TODO: Return response?

    def list_jobs(
        self,
        project_id: str,
        request_kwargs: Dict = {},
    ) -> List[Job]:
        resources = self.client.get(f"/projects/{project_id}/jobs", **request_kwargs)
        handle_errors(resources)
        return parse_raw_as(List[Job], resources.text)

    def deploy_job(
        self,
        project_id: str,
        job_input: JobInput,
        action_id: Optional[str] = None,
        wait: bool = False,
        request_kwargs: Dict = {},
    ) -> Job:
        params = {}
        if wait:
            params["wait"] = "true"
        if action_id:
            params["action_id"] = action_id
        resource = self.client.post(
            f"/projects/{project_id}/jobs",
            params=params,
            data=job_input.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(Job, resource.text)

    def list_deploy_job_actions(
        self,
        project_id: str,
        job: JobInput,
        request_kwargs: Dict = {},
    ) -> List[ResourceAction]:
        resources = self.client.post(
            f"/projects/{project_id}/jobs:deploy-actions",
            data=job.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resources)
        return parse_raw_as(List[ResourceAction], resources.text)

    def suggest_job_config(
        self,
        project_id: str,
        container_image: str,
        request_kwargs: Dict = {},
    ) -> JobInput:
        resource = self.client.get(
            f"/projects/{project_id}/jobs:suggest-config",
            params={"container_image": container_image},
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(JobInput, resource.text)

    def get_job_metadata(
        self,
        project_id: str,
        job_id: str,
        request_kwargs: Dict = {},
    ) -> Job:
        resource = self.client.get(
            f"/projects/{project_id}/jobs/{job_id}", **request_kwargs
        )
        handle_errors(resource)
        return parse_raw_as(Job, resource.text)

    def delete_job(
        self,
        project_id: str,
        job_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.delete(
            f"/projects/{project_id}/jobs/{job_id}",
            **request_kwargs,
        )
        handle_errors(response)

    def delete_jobs(
        self,
        project_id: str,
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.delete(
            f"/projects/{project_id}/jobs",
            **request_kwargs,
        )
        handle_errors(response)

    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
        request_kwargs: Dict = {},
    ) -> str:
        params = {}
        if lines:
            params["lines"] = str(lines)
        if since:
            params["since"] = since.__str__()
        response = self.client.get(
            f"/projects/{project_id}/jobs/{job_id}/logs",
            params=params,
            **request_kwargs,
        )
        handle_errors(response)
        return response.json()

    def list_job_actions(
        self,
        project_id: str,
        job_id: str,
        request_kwargs: Dict = {},
    ) -> List[ResourceAction]:
        resources = self.client.get(
            f"/projects/{project_id}/jobs/{job_id}/actions", **request_kwargs
        )
        handle_errors(resources)
        return parse_raw_as(List[ResourceAction], resources.text)

    def execute_job_action(
        self,
        project_id: str,
        job_id: str,
        action_id: str,
        action_execution: ResourceActionExecution = ResourceActionExecution(),
        request_kwargs: Dict = {},
    ) -> None:
        response = self.client.post(
            f"/projects/{project_id}/jobs/{job_id}/actions/{action_id}",
            **request_kwargs,
            data=action_execution.json(exclude_unset=True),
        )
        handle_errors(response)
        # TODO: Return response?



from fastapi import Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException, RequestValidationError
from starlette.responses import JSONResponse, Response

from contaxy.schema.auth import OAuth2Error, OAuth2ErrorDetails
from contaxy.schema.exceptions import ClientBaseError, ProblemDetails


async def handel_http_exeptions(request: Request, exc: HTTPException) -> Response:

    if isinstance(exc, OAuth2Error):
        return JSONResponse(
            status_code=exc.status_code,
            content=jsonable_encoder(OAuth2ErrorDetails(error=exc.detail)),
        )

    problem_details = ProblemDetails(code=exc.status_code, message=exc.detail)

    if isinstance(exc, ClientBaseError):
        if exc.explanation:  # type: ignore
            problem_details.explanation = exc.explanation  # type: ignore

        if exc.metadata:  # type: ignore
            problem_details.details = exc.metadata  # type: ignore

    return JSONResponse(
        status_code=exc.status_code,
        content=jsonable_encoder(
            problem_details.dict(exclude_unset=True, exclude_defaults=True)
        ),
    )


async def handle_validation_exception(
    request: Request, exc: RequestValidationError
) -> Response:
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY
    details = exc.errors()
    message = "Validation Error."
    try:
        message = details[0]["msg"]
    except Exception:
        pass

    problem_details = ProblemDetails(code=status_code, message=message, details=details)

    return JSONResponse(
        status_code=status_code,
        content=jsonable_encoder(
            problem_details.dict(exclude_unset=True, exclude_defaults=True)
        ),
    )
